import tensorflow as tf

from keras.src.backend.common import standardize_dtype
from keras.src.backend.config import floatx
from keras.src.random.seed_generator import SeedGenerator
from keras.src.random.seed_generator import draw_seed
from keras.src.random.seed_generator import make_default_seed


def _cast_seed(seed):
    # TensorFlow has a device placement issue that `Variable` must be int64
    # in `SeedGenerator`. However, all `tf.random.stateless_*` expect the seed
    # to be int32 to run with XLA.
    # This function addresses the inconsistency using `floormod`.
    # Ref: https://www.tensorflow.org/api_docs/python/tf/random
    if standardize_dtype(seed.dtype) == "int32":
        return seed
    else:
        seed = tf.cast(tf.math.floormod(seed, tf.int32.max - 1), dtype="int32")
        return seed


def normal(shape, mean=0.0, stddev=1.0, dtype=None, seed=None):
    dtype = dtype or floatx()
    seed = _cast_seed(draw_seed(seed))
    return tf.random.stateless_normal(
        shape=shape, mean=mean, stddev=stddev, dtype=dtype, seed=seed
    )


def uniform(shape, minval=0.0, maxval=1.0, dtype=None, seed=None):
    dtype = dtype or floatx()
    seed = _cast_seed(draw_seed(seed))
    return tf.random.stateless_uniform(
        shape=shape,
        minval=tf.cast(minval, dtype),
        maxval=tf.cast(maxval, dtype),
        dtype=dtype,
        seed=seed,
    )


def categorical(logits, num_samples, dtype="int64", seed=None):
    seed = _cast_seed(draw_seed(seed))
    output = tf.random.stateless_categorical(logits, num_samples, seed=seed)
    return tf.cast(output, dtype)


def randint(shape, minval, maxval, dtype="int32", seed=None):
    intermediate_dtype = dtype
    if standardize_dtype(dtype) not in ["int32", "int64"]:
        intermediate_dtype = "int64"
    seed = _cast_seed(draw_seed(seed))
    output = tf.random.stateless_uniform(
        shape=shape,
        minval=minval,
        maxval=maxval,
        dtype=intermediate_dtype,
        seed=seed,
    )
    return tf.cast(output, dtype)


def truncated_normal(shape, mean=0.0, stddev=1.0, dtype=None, seed=None):
    dtype = dtype or floatx()
    seed = _cast_seed(draw_seed(seed))
    return tf.random.stateless_truncated_normal(
        shape=shape, mean=mean, stddev=stddev, dtype=dtype, seed=seed
    )


def _get_concrete_noise_shape(inputs, noise_shape):
    if noise_shape is None:
        return tf.shape(inputs)

    concrete_inputs_shape = tf.shape(inputs)
    concrete_noise_shape = []
    for i, value in enumerate(noise_shape):
        concrete_noise_shape.append(
            concrete_inputs_shape[i] if value is None else value
        )
    return concrete_noise_shape


def dropout(inputs, rate, noise_shape=None, seed=None):
    if rate == 1.0:
        return tf.zeros_like(inputs)
    if rate == 0.0:
        return inputs
    seed = _cast_seed(draw_seed(seed))
    noise_shape = _get_concrete_noise_shape(inputs, noise_shape)
    return tf.nn.experimental.stateless_dropout(
        inputs,
        rate=rate,
        noise_shape=noise_shape,
        seed=seed,
    )


def shuffle(x, axis=0, seed=None):
    seed = _cast_seed(draw_seed(seed))
    indices = tf.argsort(
        tf.random.stateless_uniform(shape=[tf.shape(x)[axis]], seed=seed)
    )
    return tf.gather(x, indices, axis=axis)


def gamma(shape, alpha, dtype=None, seed=None):
    dtype = dtype or floatx()
    seed = _cast_seed(draw_seed(seed))
    # TODO: `tf.random.stateless_gamma` doesn't support bfloat16
    intermediate_dtype = dtype
    if standardize_dtype(dtype) == "bfloat16":
        intermediate_dtype = "float32"
    return tf.cast(
        tf.random.stateless_gamma(
            shape,
            alpha=alpha,
            dtype=intermediate_dtype,
            seed=seed,
        ),
        dtype,
    )


def binomial(shape, counts, probabilities, dtype=None, seed=None):
    dtype = dtype or floatx()
    seed = _cast_seed(draw_seed(seed))
    # TODO: `tf.random.stateless_binomial` doesn't support bfloat16
    intermediate_dtype = dtype
    if standardize_dtype(dtype) == "bfloat16":
        intermediate_dtype = "float32"
    return tf.cast(
        tf.random.stateless_binomial(
            shape=shape,
            seed=seed,
            counts=counts,
            probs=probabilities,
            output_dtype=intermediate_dtype,
        ),
        dtype,
    )


def beta(shape, alpha, beta, dtype=None, seed=None):
    dtype = dtype or floatx()
    # since tensorflow doesn't offer a beta distribution function
    # so we'll use the formula U(a,b) = (X(a) / (X(a) + Y(b)),
    # where U(a,b) is a beta-distributed random variable with
    # parameters a and b, and X(a) and Y(b) are gamma-distributed
    # random variables with parameters a and b respectively.

    # Additionally, we'll use two different seeds for our two
    # gamma random variables to prevent any unintended
    # dependencies and correlations between the generated values
    # due to the usage of same seed.
    seed_1 = _cast_seed(draw_seed(seed))
    # The choice of 12 is totally arbitrary, as we're
    # incrementing the first drawn seed by a CONSTANT to
    # ensure deterministic results.
    seed_2 = seed_1 + 12

    # TODO: `tf.random.stateless_gamma` doesn't support bfloat16
    intermediate_dtype = dtype
    if standardize_dtype(dtype) == "bfloat16":
        intermediate_dtype = "float32"
    alpha = tf.convert_to_tensor(alpha, dtype=intermediate_dtype)
    beta = tf.convert_to_tensor(beta, dtype=intermediate_dtype)

    # tensorflow's tf.random.stateless_gamma has a bit of unconventional
    # implementation of the stateless_gamma function where it checks the
    # broadcastability of alpha's shape with ONLY the RIGHTMOST dimension of
    # the specified output shape instead of considering the whole.
    # Consequently, it then results in errors for perfectly broadcastable shapes
    # such as for output shape of (2, 3) and alpha shape of (1, 3)
    # So to resolve this, we explicitly broadcast alpha and beta to shape before
    # passing them to the stateless_gamma function.
    alpha = tf.broadcast_to(alpha, shape)
    beta = tf.broadcast_to(beta, shape)

    gamma_a = tf.cast(
        tf.random.stateless_gamma(
            shape=shape, seed=seed_1, alpha=alpha, dtype=intermediate_dtype
        ),
        dtype,
    )
    gamma_b = tf.cast(
        tf.random.stateless_gamma(
            shape=shape, seed=seed_2, alpha=beta, dtype=intermediate_dtype
        ),
        dtype,
    )
    sample = gamma_a / (gamma_a + gamma_b)
    return sample



import json
import shutil
import tempfile
from pathlib import Path

import numpy as np
from absl.testing import parameterized

from keras.src import backend
from keras.src import ops
from keras.src import tree
from keras.src import utils
from keras.src.backend.common import is_float_dtype
from keras.src.backend.common import standardize_dtype
from keras.src.backend.common.global_state import clear_session
from keras.src.backend.common.keras_tensor import KerasTensor
from keras.src.losses.loss import Loss
from keras.src.models import Model
from keras.src.utils import traceback_utils


class TestCase(parameterized.TestCase):
    maxDiff = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setUp(self):
        # clear global state so that test cases are independent
        # required for the jit enabled torch tests since dynamo has
        # a global cache for guards, compiled fn, etc
        clear_session(free_memory=False)
        if traceback_utils.is_traceback_filtering_enabled():
            traceback_utils.disable_traceback_filtering()

    def get_temp_dir(self):
        temp_dir = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(temp_dir))
        return temp_dir

    def convert_to_numpy(self, x):
        if isinstance(x, np.ndarray):
            return x
        elif backend.is_tensor(x) or isinstance(x, backend.Variable):
            return backend.convert_to_numpy(x)
        return np.array(x)

    def assertAllClose(
        self,
        actual,
        desired,
        atol=1e-6,
        rtol=1e-6,
        tpu_atol=None,
        tpu_rtol=None,
        msg=None,
    ):
        if tpu_atol is not None and uses_tpu():
            atol = tpu_atol
        if tpu_rtol is not None and uses_tpu():
            rtol = tpu_rtol
        actual = self.convert_to_numpy(actual)
        desired = self.convert_to_numpy(desired)
        np.testing.assert_allclose(
            actual, desired, atol=atol, rtol=rtol, err_msg=msg
        )

    def assertNotAllClose(self, x1, x2, atol=1e-6, rtol=1e-6, msg=None):
        try:
            self.assertAllClose(x1, x2, atol=atol, rtol=rtol, msg=msg)
        except AssertionError:
            return
        msg = msg or ""
        raise AssertionError(
            f"The two values are close at all elements. \n{msg}.\nValues: {x1}"
        )

    def assertAlmostEqual(self, x1, x2, decimal=3, tpu_decimal=None, msg=None):
        if tpu_decimal is not None and uses_tpu():
            decimal = tpu_decimal
        msg = msg or ""
        x1 = self.convert_to_numpy(x1)
        x2 = self.convert_to_numpy(x2)
        np.testing.assert_almost_equal(x1, x2, decimal=decimal, err_msg=msg)

    def assertAllEqual(self, x1, x2, msg=None):
        self.assertEqual(len(x1), len(x2), msg=msg)
        for e1, e2 in zip(x1, x2):
            if isinstance(e1, (list, tuple)) or isinstance(e2, (list, tuple)):
                self.assertAllEqual(e1, e2, msg=msg)
            else:
                e1 = self.convert_to_numpy(e1)
                e2 = self.convert_to_numpy(e2)
                np.testing.assert_array_equal(e1, e2, err_msg=msg)

    def assertLen(self, iterable, expected_len, msg=None):
        self.assertEqual(len(iterable), expected_len, msg=msg)

    def assertSparse(self, x, sparse=True):
        if isinstance(x, KerasTensor):
            self.assertEqual(x.sparse, sparse)
        elif backend.backend() == "tensorflow":
            import tensorflow as tf

            if sparse:
                self.assertIsInstance(x, tf.SparseTensor)
            else:
                self.assertNotIsInstance(x, tf.SparseTensor)
        elif backend.backend() == "jax":
            import jax.experimental.sparse as jax_sparse

            if sparse:
                self.assertIsInstance(x, jax_sparse.JAXSparse)
            else:
                self.assertNotIsInstance(x, jax_sparse.JAXSparse)
        else:
            self.assertFalse(
                sparse,
                f"Backend {backend.backend()} does not support sparse tensors",
            )

    def assertRagged(self, x, ragged=True):
        if isinstance(x, KerasTensor):
            self.assertEqual(x.ragged, ragged)
        elif backend.backend() == "tensorflow":
            import tensorflow as tf

            if ragged:
                self.assertIsInstance(x, tf.RaggedTensor)
            else:
                self.assertNotIsInstance(x, tf.RaggedTensor)
        else:
            self.assertFalse(
                ragged,
                f"Backend {backend.backend()} does not support ragged tensors",
            )

    def assertDType(self, x, dtype, msg=None):
        if hasattr(x, "dtype"):
            x_dtype = backend.standardize_dtype(x.dtype)
        else:
            # If x is a python number
            x_dtype = backend.standardize_dtype(type(x))
        standardized_dtype = backend.standardize_dtype(dtype)
        default_msg = (
            "The dtype of x does not match the expected one. "
            f"Received: x.dtype={x_dtype} and dtype={dtype}"
        )
        msg = msg or default_msg
        self.assertEqual(x_dtype, standardized_dtype, msg=msg)

    def assertFileExists(self, path):
        if not Path(path).is_file():
            raise AssertionError(f"File {path} does not exist")

    def run_class_serialization_test(self, instance, custom_objects=None):
        from keras.src.saving import custom_object_scope
        from keras.src.saving import deserialize_keras_object
        from keras.src.saving import serialize_keras_object

        # get_config roundtrip
        cls = instance.__class__
        config = instance.get_config()
        config_json = to_json_with_tuples(config)
        ref_dir = dir(instance)[:]
        with custom_object_scope(custom_objects):
            revived_instance = cls.from_config(config)
        revived_config = revived_instance.get_config()
        revived_config_json = to_json_with_tuples(revived_config)
        self.assertEqual(config_json, revived_config_json)
        self.assertEqual(set(ref_dir), set(dir(revived_instance)))

        # serialization roundtrip
        serialized = serialize_keras_object(instance)
        serialized_json = to_json_with_tuples(serialized)
        with custom_object_scope(custom_objects):
            revived_instance = deserialize_keras_object(
                from_json_with_tuples(serialized_json)
            )
        revived_config = revived_instance.get_config()
        revived_config_json = to_json_with_tuples(revived_config)
        self.assertEqual(config_json, revived_config_json)
        new_dir = dir(revived_instance)[:]
        for lst in [ref_dir, new_dir]:
            if "__annotations__" in lst:
                lst.remove("__annotations__")
        self.assertEqual(set(ref_dir), set(new_dir))
        return revived_instance

    def run_layer_test(
        self,
        layer_cls,
        init_kwargs,
        input_shape=None,
        input_dtype=None,
        input_sparse=False,
        input_ragged=False,
        input_data=None,
        call_kwargs=None,
        expected_output_shape=None,
        expected_output_dtype=None,
        expected_output_sparse=False,
        expected_output_ragged=False,
        expected_output=None,
        expected_num_trainable_weights=None,
        expected_num_non_trainable_weights=None,
        expected_num_non_trainable_variables=None,
        expected_num_seed_generators=None,
        expected_num_losses=None,
        supports_masking=None,
        expected_mask_shape=None,
        custom_objects=None,
        run_training_check=True,
        run_mixed_precision_check=True,
        assert_built_after_instantiation=False,
        tpu_atol=None,
        tpu_rtol=None,
    ):
        """Run basic checks on a layer.

        Args:
            layer_cls: The class of the layer to test.
            init_kwargs: Dict of arguments to be used to
                instantiate the layer.
            input_shape: Shape tuple (or list/dict of shape tuples)
                to call the layer on.
            input_dtype: Corresponding input dtype.
            input_sparse: Whether the input is a sparse tensor (this requires
                the backend to support sparse tensors).
            input_ragged: Whether the input is a ragged tensor (this requires
                the backend to support ragged tensors).
            input_data: Tensor (or list/dict of tensors)
                to call the layer on.
            call_kwargs: Dict of arguments to use when calling the
                layer (does not include the first input tensor argument)
            expected_output_shape: Shape tuple
                (or list/dict of shape tuples)
                expected as output.
            expected_output_dtype: dtype expected as output.
            expected_output_sparse: Whether the output is expected to be sparse
                (this requires the backend to support sparse tensors).
            expected_output_ragged: Whether the output is expected to be ragged
                (this requires the backend to support ragged tensors).
            expected_output: Expected output tensor -- only
                to be specified if input_data is provided.
            expected_num_trainable_weights: Expected number
                of trainable weights of the layer once built.
            expected_num_non_trainable_weights: Expected number
                of non-trainable weights of the layer once built.
            expected_num_seed_generators: Expected number of
                SeedGenerators objects of the layer once built.
            expected_num_losses: Expected number of loss tensors
                produced when calling the layer.
            supports_masking: If True, will check that the layer
                supports masking.
            expected_mask_shape: Expected mask shape tuple
                returned by compute_mask() (only supports 1 shape).
            custom_objects: Dict of any custom objects to be
                considered during deserialization.
            run_training_check: Whether to attempt to train the layer
                (if an input shape or input data was provided).
            run_mixed_precision_check: Whether to test the layer with a mixed
                precision dtype policy.
            assert_built_after_instantiation: Whether to assert `built=True`
                after the layer's instantiation.
        """
        if input_shape is not None and input_data is not None:
            raise ValueError(
                "input_shape and input_data cannot be passed at the same time."
            )
        if expected_output_shape is not None and expected_output is not None:
            raise ValueError(
                "expected_output_shape and expected_output cannot be passed "
                "at the same time."
            )
        if expected_output is not None and input_data is None:
            raise ValueError(
                "In order to use expected_output, input_data must be provided."
            )
        if expected_mask_shape is not None and supports_masking is not True:
            raise ValueError(
                "In order to use expected_mask_shape, supports_masking "
                "must be True."
            )

        init_kwargs = init_kwargs or {}
        call_kwargs = call_kwargs or {}

        if input_shape is not None and input_dtype is not None:
            if isinstance(input_shape, tuple) and is_shape_tuple(
                input_shape[0]
            ):
                self.assertIsInstance(input_dtype, tuple)
                self.assertEqual(
                    len(input_shape),
                    len(input_dtype),
                    msg="The number of input shapes and dtypes does not match",
                )
            elif isinstance(input_shape, dict):
                self.assertIsInstance(input_dtype, dict)
                self.assertEqual(
                    set(input_shape.keys()),
                    set(input_dtype.keys()),
                    msg="The number of input shapes and dtypes does not match",
                )
            elif isinstance(input_shape, list):
                self.assertIsInstance(input_dtype, list)
                self.assertEqual(
                    len(input_shape),
                    len(input_dtype),
                    msg="The number of input shapes and dtypes does not match",
                )
            elif not isinstance(input_shape, tuple):
                raise ValueError("The type of input_shape is not supported")
        if input_shape is not None and input_dtype is None:
            input_dtype = tree.map_shape_structure(
                lambda _: "float32", input_shape
            )

        # Estimate actual number of weights, variables, seed generators if
        # expected ones not set. When using layers uses composition it should
        # build each sublayer manually.
        if input_data is not None or input_shape is not None:
            if input_data is None:
                input_data = create_eager_tensors(
                    input_shape, input_dtype, input_sparse, input_ragged
                )
            layer = layer_cls(**init_kwargs)
            if isinstance(input_data, dict):
                layer(**input_data, **call_kwargs)
            else:
                layer(input_data, **call_kwargs)

            if expected_num_trainable_weights is None:
                expected_num_trainable_weights = len(layer.trainable_weights)
            if expected_num_non_trainable_weights is None:
                expected_num_non_trainable_weights = len(
                    layer.non_trainable_weights
                )
            if expected_num_non_trainable_variables is None:
                expected_num_non_trainable_variables = len(
                    layer.non_trainable_variables
                )
            if expected_num_seed_generators is None:
                expected_num_seed_generators = len(get_seed_generators(layer))

        # Serialization test.
        layer = layer_cls(**init_kwargs)
        self.run_class_serialization_test(layer, custom_objects)

        # Basic masking test.
        if supports_masking is not None:
            self.assertEqual(
                layer.supports_masking,
                supports_masking,
                msg="Unexpected supports_masking value",
            )

        def run_build_asserts(layer):
            self.assertTrue(layer.built)
            if expected_num_trainable_weights is not None:
                self.assertLen(
                    layer.trainable_weights,
                    expected_num_trainable_weights,
                    msg="Unexpected number of trainable_weights",
                )
            if expected_num_non_trainable_weights is not None:
                self.assertLen(
                    layer.non_trainable_weights,
                    expected_num_non_trainable_weights,
                    msg="Unexpected number of non_trainable_weights",
                )
            if expected_num_non_trainable_variables is not None:
                self.assertLen(
                    layer.non_trainable_variables,
                    expected_num_non_trainable_variables,
                    msg="Unexpected number of non_trainable_variables",
                )
            if expected_num_seed_generators is not None:
                self.assertLen(
                    get_seed_generators(layer),
                    expected_num_seed_generators,
                    msg="Unexpected number of seed_generators",
                )
            if (
                backend.backend() == "torch"
                and expected_num_trainable_weights is not None
                and expected_num_non_trainable_weights is not None
                and expected_num_seed_generators is not None
            ):
                self.assertLen(
                    layer.torch_params,
                    expected_num_trainable_weights
                    + expected_num_non_trainable_weights
                    + expected_num_seed_generators,
                    msg="Unexpected number of torch_params",
                )

        def run_output_asserts(
            layer, output, eager=False, tpu_atol=None, tpu_rtol=None
        ):
            if expected_output_shape is not None:

                def verify_shape(expected_shape, x):
                    shape = x.shape
                    if len(shape) != len(expected_shape):
                        return False
                    for expected_dim, dim in zip(expected_shape, shape):
                        if expected_dim is not None and expected_dim != dim:
                            return False
                    return True

                shapes_match = tree.map_structure_up_to(
                    output, verify_shape, expected_output_shape, output
                )
                self.assertTrue(
                    all(tree.flatten(shapes_match)),
                    msg=f"Expected output shapes {expected_output_shape} but "
                    f"received {tree.map_structure(lambda x: x.shape, output)}",
                )
            if expected_output_dtype is not None:

                def verify_dtype(expected_dtype, x):
                    return expected_dtype == backend.standardize_dtype(x.dtype)

                dtypes_match = tree.map_structure(
                    verify_dtype, expected_output_dtype, output
                )
                self.assertTrue(
                    all(tree.flatten(dtypes_match)),
                    msg=f"Expected output dtypes {expected_output_dtype} but "
                    f"received {tree.map_structure(lambda x: x.dtype, output)}",
                )
            if expected_output_sparse:
                for x in tree.flatten(output):
                    self.assertSparse(x)
            if expected_output_ragged:
                for x in tree.flatten(output):
                    self.assertRagged(x)
            if eager:
                if expected_output is not None:
                    self.assertEqual(type(expected_output), type(output))
                    for ref_v, v in zip(
                        tree.flatten(expected_output), tree.flatten(output)
                    ):
                        self.assertAllClose(
                            ref_v,
                            v,
                            msg="Unexpected output value",
                            tpu_atol=tpu_atol,
                            tpu_rtol=tpu_rtol,
                        )
                if expected_num_losses is not None:
                    self.assertLen(layer.losses, expected_num_losses)

        def run_training_step(layer, input_data, output_data):
            class TestModel(Model):
                def __init__(self, layer):
                    super().__init__()
                    self.layer = layer

                def call(self, x, training=False):
                    return self.layer(x, training=training)

            model = TestModel(layer)

            data = (input_data, output_data)
            if backend.backend() == "torch":
                data = tree.map_structure(backend.convert_to_numpy, data)

            def data_generator():
                while True:
                    yield data

            # Single op loss to avoid compilation issues with ragged / sparse.
            class TestLoss(Loss):
                def __call__(self, y_true, y_pred, sample_weight=None):
                    return ops.sum(y_pred)

            # test the "default" path for each backend by setting
            # jit_compile="auto".
            # for tensorflow and jax backends auto is jitted
            # Note that tensorflow cannot be jitted with sparse tensors
            # for torch backend auto is eager
            #
            # NB: for torch, jit_compile=True turns on torchdynamo
            #  which may not always succeed in tracing depending
            #  on the model. Run your program with these env vars
            #  to get debug traces of dynamo:
            #    TORCH_LOGS="+dynamo"
            #    TORCHDYNAMO_VERBOSE=1
            #    TORCHDYNAMO_REPORT_GUARD_FAILURES=1
            jit_compile = "auto"
            if backend.backend() == "tensorflow" and input_sparse:
                jit_compile = False
            model.compile(
                optimizer="sgd", loss=TestLoss(), jit_compile=jit_compile
            )
            model.fit(data_generator(), steps_per_epoch=1, verbose=0)

        # Build test.
        if input_data is not None or input_shape is not None:
            if input_shape is None:
                build_shape = tree.map_structure(
                    lambda x: ops.shape(x), input_data
                )
            else:
                build_shape = input_shape
            layer = layer_cls(**init_kwargs)
            if isinstance(build_shape, dict):
                layer.build(**build_shape)
            else:
                layer.build(build_shape)
            run_build_asserts(layer)

            # Symbolic call test.
            if input_shape is None:
                keras_tensor_inputs = tree.map_structure(
                    lambda x: create_keras_tensors(
                        ops.shape(x), x.dtype, input_sparse, input_ragged
                    ),
                    input_data,
                )
            else:
                keras_tensor_inputs = create_keras_tensors(
                    input_shape, input_dtype, input_sparse, input_ragged
                )
            layer = layer_cls(**init_kwargs)
            if isinstance(keras_tensor_inputs, dict):
                keras_tensor_outputs = layer(
                    **keras_tensor_inputs, **call_kwargs
                )
            else:
                keras_tensor_outputs = layer(keras_tensor_inputs, **call_kwargs)
            run_build_asserts(layer)
            run_output_asserts(layer, keras_tensor_outputs, eager=False)

            if expected_mask_shape is not None:
                output_mask = layer.compute_mask(keras_tensor_inputs)
                self.assertEqual(expected_mask_shape, output_mask.shape)

            # The stateless layers should be built after instantiation.
            if assert_built_after_instantiation:
                layer = layer_cls(**init_kwargs)
                self.assertTrue(
                    layer.built,
                    msg=(
                        f"{type(layer)} is stateless, so it should be built "
                        "after instantiation."
                    ),
                )

                # Ensure that the subclass layer doesn't mark itself as built
                # when `build` is overridden.

                class ModifiedBuildLayer(layer_cls):
                    def build(self, *args, **kwargs):
                        pass

                layer = ModifiedBuildLayer(**init_kwargs)
                self.assertFalse(
                    layer.built,
                    msg=(
                        f"The `build` of {type(layer)} is overriden, so it "
                        "should not be built after instantiation."
                    ),
                )

        # Eager call test and compiled training test.
        if input_data is not None or input_shape is not None:
            if input_data is None:
                input_data = create_eager_tensors(
                    input_shape, input_dtype, input_sparse
                )
            layer = layer_cls(**init_kwargs)
            if isinstance(input_data, dict):
                output_data = layer(**input_data, **call_kwargs)
            else:
                output_data = layer(input_data, **call_kwargs)
            run_output_asserts(
                layer,
                output_data,
                eager=True,
                tpu_atol=tpu_atol,
                tpu_rtol=tpu_rtol,
            )

            if run_training_check and backend.backend() not in (
                "numpy",
                "openvino",
            ):
                run_training_step(layer, input_data, output_data)

            # Never test mixed precision on torch CPU. Torch lacks support.
            if run_mixed_precision_check and backend.backend() == "torch":
                import torch

                run_mixed_precision_check = torch.cuda.is_available()

            if run_mixed_precision_check:
                layer = layer_cls(**{**init_kwargs, "dtype": "mixed_float16"})
                input_spec = tree.map_structure(
                    lambda spec: KerasTensor(
                        spec.shape,
                        dtype=(
                            layer.compute_dtype
                            if layer.autocast
                            and backend.is_float_dtype(spec.dtype)
                            else spec.dtype
                        ),
                    ),
                    keras_tensor_inputs,
                )
                if isinstance(input_data, dict):
                    output_data = layer(**input_data, **call_kwargs)
                    output_spec = layer.compute_output_spec(**input_spec)
                else:
                    output_data = layer(input_data, **call_kwargs)
                    output_spec = layer.compute_output_spec(input_spec)
                for tensor, spec in zip(
                    tree.flatten(output_data), tree.flatten(output_spec)
                ):
                    dtype = standardize_dtype(tensor.dtype)
                    self.assertEqual(
                        dtype,
                        spec.dtype,
                        f"expected output dtype {spec.dtype}, got {dtype}",
                    )
                for weight in layer.weights:
                    dtype = standardize_dtype(weight.dtype)
                    if is_float_dtype(dtype):
                        self.assertEqual(dtype, "float32")


def _jax_uses(device_type):
    import jax

    return jax.default_backend() == device_type


def _tensorflow_uses(device_type):
    import tensorflow as tf

    return len(tf.config.list_physical_devices(device_type.upper())) > 0


def _torch_uses(device_type):
    if device_type == "gpu":
        from keras.src.backend.torch.core import get_device

        return get_device() == "cuda"
    return device_type == "cpu"


def uses_gpu():
    if not hasattr(uses_gpu, "_value"):
        if backend.backend() == "tensorflow":
            uses_gpu._value = _tensorflow_uses("gpu")
        elif backend.backend() == "jax":
            uses_gpu._value = _jax_uses("gpu")
        elif backend.backend() == "torch":
            uses_gpu._value = _torch_uses("gpu")
        else:
            uses_gpu._value = False
    return uses_gpu._value


def uses_tpu():
    if not hasattr(uses_tpu, "_value"):
        if backend.backend() == "tensorflow":
            uses_tpu._value = _tensorflow_uses("tpu")
        elif backend.backend() == "jax":
            uses_tpu._value = _jax_uses("tpu")
        else:
            uses_tpu._value = False
    return uses_tpu._value


def jax_uses_gpu():
    return backend.backend() == "jax" and uses_gpu()


def jax_uses_tpu():
    return backend.backend() == "jax" and uses_tpu()


def tensorflow_uses_gpu():
    return backend.backend() == "tensorflow" and uses_gpu()


def tensorflow_uses_tpu():
    return backend.backend() == "tensorflow" and uses_tpu()


def torch_uses_gpu():
    return backend.backend() == "torch" and uses_gpu()


def create_keras_tensors(input_shape, dtype, sparse, ragged):
    if isinstance(input_shape, dict):
        return {
            utils.removesuffix(k, "_shape"): KerasTensor(
                v, dtype=dtype[k], sparse=sparse, ragged=ragged
            )
            for k, v in input_shape.items()
        }
    return map_shape_dtype_structure(
        lambda shape, dt: KerasTensor(
            shape, dtype=dt, sparse=sparse, ragged=ragged
        ),
        input_shape,
        dtype,
    )


def create_eager_tensors(input_shape, dtype, sparse, ragged):
    from keras.src.backend import random

    if set(tree.flatten(dtype)).difference(
        [
            "float16",
            "float32",
            "float64",
            "int8",
            "uint8",
            "int16",
            "uint16",
            "int32",
            "uint32",
            "int64",
            "uint64",
        ]
    ):
        raise ValueError(
            "dtype must be a standard float or int dtype. "
            f"Received: dtype={dtype}"
        )

    if sparse:
        if backend.backend() == "tensorflow":
            import tensorflow as tf

            def create_fn(shape, dt):
                rng = np.random.default_rng(0)
                x = (4 * rng.standard_normal(shape)).astype(dt)
                x = np.multiply(x, rng.random(shape) < 0.7)
                return tf.sparse.from_dense(x)

        elif backend.backend() == "jax":
            import jax.experimental.sparse as jax_sparse

            def create_fn(shape, dt):
                rng = np.random.default_rng(0)
                x = (4 * rng.standard_normal(shape)).astype(dt)
                x = np.multiply(x, rng.random(shape) < 0.7)
                return jax_sparse.BCOO.fromdense(x, n_batch=1)

        else:
            raise ValueError(
                f"Sparse is unsupported with backend {backend.backend()}"
            )

    elif ragged:
        if backend.backend() == "tensorflow":
            import tensorflow as tf

            def create_fn(shape, dt):
                rng = np.random.default_rng(0)
                x = (4 * rng.standard_normal(shape)).astype(dt)
                x = np.multiply(x, rng.random(shape) < 0.7)
                return tf.RaggedTensor.from_tensor(x, padding=0)

        else:
            raise ValueError(
                f"Ragged is unsupported with backend {backend.backend()}"
            )

    else:

        def create_fn(shape, dt):
            return ops.cast(
                random.uniform(shape, dtype="float32") * 3, dtype=dt
            )

    if isinstance(input_shape, dict):
        return {
            utils.removesuffix(k, "_shape"): create_fn(v, dtype[k])
            for k, v in input_shape.items()
        }
    return map_shape_dtype_structure(create_fn, input_shape, dtype)


def is_shape_tuple(x):
    return isinstance(x, (list, tuple)) and all(
        isinstance(e, (int, type(None))) for e in x
    )


def map_shape_dtype_structure(fn, shape, dtype):
    """Variant of tree.map_structure that operates on shape tuples."""
    if is_shape_tuple(shape):
        return fn(tuple(shape), dtype)
    if isinstance(shape, list):
        return [
            map_shape_dtype_structure(fn, s, d) for s, d in zip(shape, dtype)
        ]
    if isinstance(shape, tuple):
        return tuple(
            map_shape_dtype_structure(fn, s, d) for s, d in zip(shape, dtype)
        )
    if isinstance(shape, dict):
        return {
            k: map_shape_dtype_structure(fn, v, dtype[k])
            for k, v in shape.items()
        }
    else:
        raise ValueError(
            f"Cannot map function to unknown objects {shape} and {dtype}"
        )


def get_seed_generators(layer):
    """Get a List of all seed generators in the layer recursively."""
    seed_generators = []
    seen_ids = set()
    for sublayer in layer._flatten_layers(True, True):
        for sg in sublayer._seed_generators:
            if id(sg) not in seen_ids:
                seed_generators.append(sg)
                seen_ids.add(id(sg))
    return seed_generators


def to_json_with_tuples(value):
    def _tuple_encode(obj):
        if isinstance(obj, tuple):
            return {"__class__": "tuple", "__value__": list(obj)}
        if isinstance(obj, list):
            return [_tuple_encode(e) for e in obj]
        if isinstance(obj, dict):
            return {key: _tuple_encode(value) for key, value in obj.items()}
        return obj

    class _PreserveTupleJsonEncoder(json.JSONEncoder):
        def encode(self, obj):
            obj = _tuple_encode(obj)
            return super().encode(obj)

    return _PreserveTupleJsonEncoder(sort_keys=True, indent=4).encode(value)


def from_json_with_tuples(value):
    def _tuple_decode(obj):
        if not isinstance(obj, dict):
            return obj
        if "__class__" not in obj or "__value__" not in obj:
            return obj
        return tuple(obj["__value__"])

    return json.loads(value, object_hook=_tuple_decode)



import openvino.opset15 as ov_opset
from openvino import Model
from openvino import Type

from keras.src import tree
from keras.src.backend.openvino.core import OpenVINOKerasTensor
from keras.src.backend.openvino.core import get_ov_output


def rnn(
    step_function,
    inputs,
    initial_states,
    go_backwards=False,
    mask=None,
    constants=None,
    unroll=False,
    input_length=None,
    time_major=False,
    zero_output_for_mask=False,
    return_all_outputs=True,
):
    def swap_batch_timestep(input_t):
        axes = list(range(len(input_t.shape)))
        axes[0], axes[1] = 1, 0
        perm_const = ov_opset.constant(axes, dtype=Type.i32).output(0)
        input_ov = get_ov_output(input_t)
        return OpenVINOKerasTensor(
            ov_opset.transpose(input_ov, perm_const).output(0)
        )

    if not time_major:
        inputs = tree.map_structure(swap_batch_timestep, inputs)
        if mask is not None:
            mask = swap_batch_timestep(mask)
    flattened_inputs = tree.flatten(inputs)
    flattened_states = tree.flatten(initial_states)
    flattened_constants = tree.flatten(constants) if constants else []
    input_0 = flattened_inputs[0]
    input_0_ov = get_ov_output(input_0)
    input_shape = ov_opset.shape_of(input_0_ov, Type.i32).output(0)
    time_steps = ov_opset.gather(
        input_shape,
        ov_opset.constant([0], dtype=Type.i32).output(0),
        ov_opset.constant(0, dtype=Type.i32).output(0),
    ).output(0)
    time_steps = ov_opset.squeeze(
        time_steps, ov_opset.constant([0], dtype=Type.i32).output(0)
    ).output(0)
    if mask is None and input_length is not None:
        input_len_ov = get_ov_output(input_length)
        if input_len_ov.get_partial_shape().rank.get_length() == 1:
            indices = ov_opset.range(
                ov_opset.constant(0, dtype=Type.i32).output(0),
                time_steps,
                ov_opset.constant(1, dtype=Type.i32).output(0),
                output_type=Type.i32,
            ).output(0)
            indices = ov_opset.unsqueeze(
                indices, ov_opset.constant([1], dtype=Type.i32).output(0)
            ).output(0)
            input_len_casted = ov_opset.convert(input_len_ov, Type.i32).output(
                0
            )
            input_len_expanded = ov_opset.unsqueeze(
                input_len_casted,
                ov_opset.constant([0], dtype=Type.i32).output(0),
            ).output(0)
            mask_bool = ov_opset.less(indices, input_len_expanded).output(0)
            mask = OpenVINOKerasTensor(mask_bool)
    if mask is not None:
        mask_ov = get_ov_output(mask)
        if mask_ov.get_element_type() != Type.boolean:
            mask_ov = ov_opset.convert(mask_ov, Type.boolean).output(0)
        pshape = mask_ov.get_partial_shape()
        rank = pshape.rank.get_length()
        if rank == 2:
            mask_ov = ov_opset.unsqueeze(
                mask_ov, ov_opset.constant([-1], dtype=Type.i32).output(0)
            ).output(0)
        mask = OpenVINOKerasTensor(mask_ov)
    if go_backwards:

        def reverse_time(x):
            x_ov = get_ov_output(x)
            start = ov_opset.constant([0], dtype=Type.i32).output(0)
            idx = ov_opset.range(
                ov_opset.subtract(
                    time_steps, ov_opset.constant(1, dtype=Type.i32).output(0)
                ).output(0),
                ov_opset.constant(-1, dtype=Type.i32).output(0),
                ov_opset.constant(-1, dtype=Type.i32).output(0),
                output_type=Type.i32,
            ).output(0)
            return OpenVINOKerasTensor(
                ov_opset.gather(x_ov, idx, start).output(0)
            )

        inputs = tree.map_structure(reverse_time, inputs)
        if mask is not None:
            mask = reverse_time(mask)
        flattened_inputs = tree.flatten(inputs)

    def _slice_at_0(x):
        x_ov = get_ov_output(x)
        slice_0 = ov_opset.gather(
            x_ov,
            ov_opset.constant(0, dtype=Type.i32).output(0),
            ov_opset.constant(0, dtype=Type.i32).output(0),
        ).output(0)
        return OpenVINOKerasTensor(slice_0)

    inputs_0 = tree.map_structure(_slice_at_0, inputs)
    output_0, _ = step_function(
        inputs_0, tuple(initial_states) + tuple(constants or [])
    )
    flattened_output_0 = tree.flatten(output_0)
    last_output_states = []
    for out in flattened_output_0:
        out_ov = get_ov_output(out)
        shape = ov_opset.shape_of(out_ov, Type.i32).output(0)
        dtype = out_ov.get_element_type()
        zeros = ov_opset.broadcast(
            ov_opset.constant(0, dtype).output(0), shape
        ).output(0)
        last_output_states.append(OpenVINOKerasTensor(zeros))
    params = []
    sliced_inputs_params = []
    for inp in flattened_inputs:
        inp_ov = get_ov_output(inp)
        pshape = inp_ov.get_partial_shape()
        if pshape.rank.is_static:
            new_shape = [1] + list(pshape)[1:]
        else:
            new_shape = None
        param = ov_opset.parameter(new_shape, inp_ov.get_element_type())
        sliced_inputs_params.append(param)
        params.append(param)
    sliced_mask_params = []
    if mask is not None:
        mask_ov = get_ov_output(mask)
        pshape = mask_ov.get_partial_shape()
        new_shape = [1] + list(pshape)[1:] if pshape.rank.is_static else None
        param = ov_opset.parameter(new_shape, mask_ov.get_element_type())
        sliced_mask_params.append(param)
        params.append(param)
    merged_states_params = []
    for st in flattened_states:
        st_ov = get_ov_output(st)
        param = ov_opset.parameter(
            st_ov.get_partial_shape(), st_ov.get_element_type()
        )
        merged_states_params.append(param)
        params.append(param)
    last_output_params = []
    for lo in last_output_states:
        lo_ov = get_ov_output(lo)
        param = ov_opset.parameter(
            lo_ov.get_partial_shape(), lo_ov.get_element_type()
        )
        last_output_params.append(param)
        params.append(param)
    constants_params = []
    for c in flattened_constants:
        c_ov = get_ov_output(c)
        param = ov_opset.parameter(
            c_ov.get_partial_shape(), c_ov.get_element_type()
        )
        constants_params.append(param)
        params.append(param)
    sliced_inputs_t = [
        OpenVINOKerasTensor(
            ov_opset.squeeze(
                p.output(0),
                ov_opset.constant([0], dtype=Type.i32).output(0),
            ).output(0)
        )
        for p in sliced_inputs_params
    ]
    merged_states_t = [
        OpenVINOKerasTensor(p.output(0)) for p in merged_states_params
    ]
    constants_t = [OpenVINOKerasTensor(p.output(0)) for p in constants_params]

    packed_inputs = tree.pack_sequence_as(inputs, sliced_inputs_t)
    packed_states = tree.pack_sequence_as(initial_states, merged_states_t)
    step_output, step_new_states = step_function(
        packed_inputs, tuple(packed_states) + tuple(constants_t)
    )
    flat_step_output = tree.flatten(step_output)
    flat_step_new_states = tree.flatten(step_new_states)
    final_output_list = []
    final_states_list = []
    final_last_output_list = []
    if mask is not None:
        mask_t = ov_opset.squeeze(
            sliced_mask_params[0].output(0),
            ov_opset.constant([0], dtype=Type.i32).output(0),
        ).output(0)
        for i, (new_st, old_st) in enumerate(
            zip(flat_step_new_states, merged_states_t)
        ):
            new_st_ov = get_ov_output(new_st)
            old_st_ov = get_ov_output(old_st)
            res = ov_opset.select(mask_t, new_st_ov, old_st_ov).output(0)
            final_states_list.append(res)
        for i, (new_out, old_last_out) in enumerate(
            zip(flat_step_output, last_output_params)
        ):
            new_out_ov = get_ov_output(new_out)
            old_last_out_ov = old_last_out.output(0)
            last_out_res = ov_opset.select(
                mask_t, new_out_ov, old_last_out_ov
            ).output(0)
            final_last_output_list.append(last_out_res)
            if zero_output_for_mask:
                zero = ov_opset.broadcast(
                    ov_opset.constant(0, new_out_ov.get_element_type()).output(
                        0
                    ),
                    ov_opset.shape_of(new_out_ov, Type.i32).output(0),
                ).output(0)
                seq_out_res = ov_opset.select(mask_t, new_out_ov, zero).output(
                    0
                )
            else:
                seq_out_res = last_out_res
            final_output_list.append(seq_out_res)
    else:
        final_states_list = [get_ov_output(x) for x in flat_step_new_states]
        final_output_list = [get_ov_output(x) for x in flat_step_output]
        final_last_output_list = [get_ov_output(x) for x in flat_step_output]
    unsq_ax = ov_opset.constant([0], dtype=Type.i32).output(0)
    final_output_list = [
        ov_opset.unsqueeze(x, unsq_ax).output(0) for x in final_output_list
    ]
    cond_const = ov_opset.constant(True, Type.boolean).output(0)
    results = (
        [cond_const]
        + final_states_list
        + final_last_output_list
        + final_output_list
    )
    body_model = Model(results, params)
    exec_cond_in = ov_opset.constant(True, Type.boolean).output(0)
    loop = ov_opset.loop(time_steps, exec_cond_in)
    loop.set_function(body_model)
    loop.set_special_body_ports([-1, 0])
    for param, inp in zip(sliced_inputs_params, flattened_inputs):
        loop.set_sliced_input(param, get_ov_output(inp), 0, 1, 1, -1, 0)
    if mask is not None:
        mask_ov = get_ov_output(mask)
        loop.set_sliced_input(sliced_mask_params[0], mask_ov, 0, 1, 1, -1, 0)
    current_res_idx = 1
    for param, init in zip(merged_states_params, flattened_states):
        loop.set_merged_input(
            param, get_ov_output(init), results[current_res_idx]
        )
        current_res_idx += 1
    final_last_outputs_res = []
    for param, init in zip(last_output_params, last_output_states):
        loop.set_merged_input(
            param, get_ov_output(init), results[current_res_idx]
        )
        final_last_outputs_res.append(results[current_res_idx])
        current_res_idx += 1
    for param, val in zip(constants_params, flattened_constants):
        loop.set_invariant_input(param, get_ov_output(val))
    loop_outputs = []
    for _ in final_output_list:
        out = loop.get_concatenated_slices(
            results[current_res_idx], 0, 1, 1, -1, 0
        )
        loop_outputs.append(OpenVINOKerasTensor(out))
        current_res_idx += 1
    loop_final_states = []
    st_res_idx = 1
    for _ in flattened_states:
        out = loop.get_iter_value(results[st_res_idx], -1)
        loop_final_states.append(OpenVINOKerasTensor(out))
        st_res_idx += 1
    lo_res_idx = st_res_idx
    loop_final_last_outputs = []
    for _ in last_output_states:
        out = loop.get_iter_value(results[lo_res_idx], -1)
        loop_final_last_outputs.append(OpenVINOKerasTensor(out))
        lo_res_idx += 1
    outputs = tree.pack_sequence_as(output_0, loop_outputs)
    new_states = tree.pack_sequence_as(initial_states, loop_final_states)
    last_output = tree.pack_sequence_as(output_0, loop_final_last_outputs)
    if not time_major:
        outputs = tree.map_structure(swap_batch_timestep, outputs)
    return last_output, outputs, new_states


def _reorder_gates(x_ov, from_order, to_order, axis):
    """Reorder gate slices of a tensor along `axis`.

    `from_order` and `to_order` are lists of single-char gate names, e.g.
    from_order=['i','f','c','o'], to_order=['f','i','c','o'].
    The tensor dimension along `axis` must be divisible by len(from_order).
    """
    n_gates = len(from_order)
    axis_const = ov_opset.constant(axis, dtype=Type.i32).output(0)
    chunks = ov_opset.split(x_ov, axis_const, n_gates).outputs()
    gate_map = {g: chunks[i] for i, g in enumerate(from_order)}
    reordered = [gate_map[g] for g in to_order]
    return ov_opset.concat(reordered, axis=axis).output(0)


def _seq_lengths(inputs_ov):
    """Return int32 sequence-length tensor [batch] equal to full time steps."""
    input_shape = ov_opset.shape_of(inputs_ov, Type.i32).output(0)
    batch_size = ov_opset.gather(
        input_shape,
        ov_opset.constant([0], dtype=Type.i32).output(0),
        ov_opset.constant(0, dtype=Type.i32).output(0),
    ).output(0)
    time_steps = ov_opset.gather(
        input_shape,
        ov_opset.constant([1], dtype=Type.i32).output(0),
        ov_opset.constant(0, dtype=Type.i32).output(0),
    ).output(0)
    return ov_opset.broadcast(time_steps, batch_size).output(0)


def lstm(
    inputs,
    initial_h,
    initial_c,
    mask,
    kernel,
    recurrent_kernel,
    bias,
    activation,
    recurrent_activation,
    return_sequences=False,
    go_backwards=False,
    unroll=False,
):
    act_name = getattr(activation, "__name__", None)
    rec_act_name = getattr(recurrent_activation, "__name__", None)
    if not (
        act_name == "tanh"
        and rec_act_name == "sigmoid"
        and not unroll
        and bias is not None
        and mask is None
    ):
        raise NotImplementedError

    inputs_ov = get_ov_output(inputs)
    initial_h_ov = get_ov_output(initial_h)
    initial_c_ov = get_ov_output(initial_c)
    kernel_ov = get_ov_output(kernel)
    recurrent_kernel_ov = get_ov_output(recurrent_kernel)
    bias_ov = get_ov_output(bias)

    hidden_size = recurrent_kernel_ov.get_partial_shape()[0].get_length()

    kt = ov_opset.transpose(
        kernel_ov,
        ov_opset.constant([1, 0], dtype=Type.i32).output(0),
    ).output(0)
    w = _reorder_gates(kt, ["i", "f", "c", "o"], ["f", "i", "c", "o"], axis=0)
    w = ov_opset.unsqueeze(
        w, ov_opset.constant([0], dtype=Type.i32).output(0)
    ).output(0)

    rt = ov_opset.transpose(
        recurrent_kernel_ov,
        ov_opset.constant([1, 0], dtype=Type.i32).output(0),
    ).output(0)
    r = _reorder_gates(rt, ["i", "f", "c", "o"], ["f", "i", "c", "o"], axis=0)
    r = ov_opset.unsqueeze(
        r, ov_opset.constant([0], dtype=Type.i32).output(0)
    ).output(0)

    b = _reorder_gates(
        bias_ov, ["i", "f", "c", "o"], ["f", "i", "c", "o"], axis=0
    )
    b = ov_opset.unsqueeze(
        b, ov_opset.constant([0], dtype=Type.i32).output(0)
    ).output(0)

    h0 = ov_opset.unsqueeze(
        initial_h_ov, ov_opset.constant([1], dtype=Type.i32).output(0)
    ).output(0)
    c0 = ov_opset.unsqueeze(
        initial_c_ov, ov_opset.constant([1], dtype=Type.i32).output(0)
    ).output(0)

    seq_lens = _seq_lengths(inputs_ov)
    direction = "reverse" if go_backwards else "forward"

    lstm_out = ov_opset.lstm_sequence(
        inputs_ov, h0, c0, seq_lens, w, r, b, hidden_size, direction
    )
    dir_axis = ov_opset.constant([1], dtype=Type.i32).output(0)
    all_outputs = ov_opset.squeeze(lstm_out.output(0), dir_axis).output(0)
    h_n = ov_opset.squeeze(lstm_out.output(1), dir_axis).output(0)
    c_n = ov_opset.squeeze(lstm_out.output(2), dir_axis).output(0)

    if go_backwards:
        shape = ov_opset.shape_of(all_outputs, Type.i32).output(0)
        time_len = ov_opset.gather(
            shape,
            ov_opset.constant(1, dtype=Type.i32).output(0),
            ov_opset.constant(0, dtype=Type.i32).output(0),
        ).output(0)
        idx = ov_opset.range(
            ov_opset.subtract(
                time_len, ov_opset.constant(1, dtype=Type.i32).output(0)
            ).output(0),
            ov_opset.constant(-1, dtype=Type.i32).output(0),
            ov_opset.constant(-1, dtype=Type.i32).output(0),
            output_type=Type.i32,
        ).output(0)
        all_outputs = ov_opset.gather(
            all_outputs,
            idx,
            ov_opset.constant(1, dtype=Type.i32).output(0),
        ).output(0)

    return (
        OpenVINOKerasTensor(h_n),
        OpenVINOKerasTensor(all_outputs),
        [OpenVINOKerasTensor(h_n), OpenVINOKerasTensor(c_n)],
    )


def gru(
    inputs,
    initial_state,
    mask,
    kernel,
    recurrent_kernel,
    bias,
    activation,
    recurrent_activation,
    return_sequences=False,
    go_backwards=False,
    unroll=False,
    reset_after=True,
):
    act_name = getattr(activation, "__name__", None)
    rec_act_name = getattr(recurrent_activation, "__name__", None)
    if not (
        act_name == "tanh"
        and rec_act_name == "sigmoid"
        and not unroll
        and bias is not None
        and reset_after
        and mask is None
    ):
        raise NotImplementedError

    inputs_ov = get_ov_output(inputs)
    initial_state_ov = get_ov_output(initial_state)
    kernel_ov = get_ov_output(kernel)
    recurrent_kernel_ov = get_ov_output(recurrent_kernel)
    bias_ov = get_ov_output(bias)

    hidden_size = recurrent_kernel_ov.get_partial_shape()[0].get_length()

    w = ov_opset.transpose(
        kernel_ov,
        ov_opset.constant([1, 0], dtype=Type.i32).output(0),
    ).output(0)
    w = ov_opset.unsqueeze(
        w, ov_opset.constant([0], dtype=Type.i32).output(0)
    ).output(0)

    r = ov_opset.transpose(
        recurrent_kernel_ov,
        ov_opset.constant([1, 0], dtype=Type.i32).output(0),
    ).output(0)
    r = ov_opset.unsqueeze(
        r, ov_opset.constant([0], dtype=Type.i32).output(0)
    ).output(0)

    # Keras bias [2, 3*units]: row 0 = input biases [b_z, b_r, b_h],
    # row 1 = recurrent biases [rb_z, rb_r, rb_h].
    # OV gru_sequence (linear_before_reset=True) wants B [1, 4*units]:
    # [b_z+rb_z, b_r+rb_r, b_h, rb_h]
    ax = ov_opset.constant(0, dtype=Type.i32).output(0)
    b_input = ov_opset.gather(
        bias_ov, ov_opset.constant(0, dtype=Type.i32).output(0), ax
    ).output(0)
    b_recur = ov_opset.gather(
        bias_ov, ov_opset.constant(1, dtype=Type.i32).output(0), ax
    ).output(0)
    split_ax = ov_opset.constant(0, dtype=Type.i32).output(0)
    b_in_parts = ov_opset.split(b_input, split_ax, 3).outputs()
    b_rc_parts = ov_opset.split(b_recur, split_ax, 3).outputs()
    b_z = ov_opset.add(b_in_parts[0], b_rc_parts[0]).output(0)
    b_r = ov_opset.add(b_in_parts[1], b_rc_parts[1]).output(0)
    b_h = b_in_parts[2]
    rb_h = b_rc_parts[2]
    b = ov_opset.concat([b_z, b_r, b_h, rb_h], axis=0).output(0)
    b = ov_opset.unsqueeze(
        b, ov_opset.constant([0], dtype=Type.i32).output(0)
    ).output(0)

    h0 = ov_opset.unsqueeze(
        initial_state_ov, ov_opset.constant([1], dtype=Type.i32).output(0)
    ).output(0)

    seq_lens = _seq_lengths(inputs_ov)
    direction = "reverse" if go_backwards else "forward"

    gru_out = ov_opset.gru_sequence(
        inputs_ov,
        h0,
        seq_lens,
        w,
        r,
        b,
        hidden_size,
        direction,
        linear_before_reset=True,
    )
    dir_axis = ov_opset.constant([1], dtype=Type.i32).output(0)
    all_outputs = ov_opset.squeeze(gru_out.output(0), dir_axis).output(0)
    h_n = ov_opset.squeeze(gru_out.output(1), dir_axis).output(0)

    if go_backwards:
        # OV direction="reverse" outputs Y in original time order
        # (Y[0]=fully-accumulated state). Keras go_backwards expects
        # Y[0]=state after first reversed step. Flip time axis to match.
        shape = ov_opset.shape_of(all_outputs, Type.i32).output(0)
        time_len = ov_opset.gather(
            shape,
            ov_opset.constant(1, dtype=Type.i32).output(0),
            ov_opset.constant(0, dtype=Type.i32).output(0),
        ).output(0)
        idx = ov_opset.range(
            ov_opset.subtract(
                time_len, ov_opset.constant(1, dtype=Type.i32).output(0)
            ).output(0),
            ov_opset.constant(-1, dtype=Type.i32).output(0),
            ov_opset.constant(-1, dtype=Type.i32).output(0),
            output_type=Type.i32,
        ).output(0)
        all_outputs = ov_opset.gather(
            all_outputs,
            idx,
            ov_opset.constant(1, dtype=Type.i32).output(0),
        ).output(0)

    return (
        OpenVINOKerasTensor(h_n),
        OpenVINOKerasTensor(all_outputs),
        [OpenVINOKerasTensor(h_n)],
    )


def cudnn_ok(*args, **kwargs):
    return False



from keras.src import backend
from keras.src import ops

DTYPE_TO_SIZE = {
    **{f"float{i}": i for i in (16, 32, 64)},
    **{f"int{i}": i for i in (8, 16, 32, 64)},
    **{f"uint{i}": i for i in (8, 16, 32, 64)},
    "bfloat16": 16,
    "bool": 1,
}


def dtype_size(dtype):
    size = DTYPE_TO_SIZE.get(dtype, None)
    if size is None:
        raise ValueError(f"Invalid dtype: {dtype}")
    return size


def is_float(dtype):
    return "float" in dtype


def cast_to_common_dtype(tensors):
    """Cast a list of tensors to a common dtype.

    If any tensor is floating-point, all tensors are cast to a common
    floating-point dtype with sufficient precision. The promotion follows
    the highest precision floating dtype present, with special handling
    for mixed `float16` and `bfloat16`, which are promoted to `float32`.

    If no floating-point tensors are present, tensors are returned unchanged.

    Args:
        tensors: A list of tensors.

    Returns:
        List of tensors cast to a common dtype when needed.
    """
    highest_float = None
    highest_float_size = -1

    seen_float16 = False
    seen_bfloat16 = False

    for x in tensors:
        dtype = backend.standardize_dtype(x.dtype)

        if is_float(dtype):
            if dtype == "float16":
                seen_float16 = True
            elif dtype == "bfloat16":
                seen_bfloat16 = True

            if highest_float is None or dtype_size(dtype) > highest_float_size:
                highest_float = dtype
                highest_float_size = dtype_size(dtype)

    # Promote mixed float16 + bfloat16 to float32
    # Do not downgrade if higher precision already found (e.g., float64)
    if seen_float16 and seen_bfloat16 and highest_float_size < 32:
        highest_float = "float32"

    if highest_float:
        tensors = [ops.cast(x, highest_float) for x in tensors]

    return tensors



import numpy as np

try:
    import sklearn
    from sklearn.base import BaseEstimator
    from sklearn.base import TransformerMixin
except ImportError:
    sklearn = None

    class BaseEstimator:
        pass

    class TransformerMixin:
        pass


def assert_sklearn_installed(symbol_name):
    if sklearn is None:
        raise ImportError(
            f"{symbol_name} requires `scikit-learn` to be installed. "
            "Run `pip install scikit-learn` to install it."
        )


def _check_model(model):
    """Check whether the model need sto be compiled."""
    # compile model if user gave us an un-compiled model
    if not model.compiled or not model.loss or not model.optimizer:
        raise RuntimeError(
            "Given model needs to be compiled, and have a loss "
            "and an optimizer."
        )


class TargetReshaper(TransformerMixin, BaseEstimator):
    """Convert 1D targets to 2D and back.

    For use in pipelines with transformers that only accept
    2D inputs, like OneHotEncoder and OrdinalEncoder.

    Attributes:
        ndim_ : int
            Dimensions of y that the transformer was trained on.
    """

    def fit(self, y):
        """Fit the transformer to a target y.

        Returns:
            TargetReshaper
                A reference to the current instance of TargetReshaper.
        """
        self.ndim_ = y.ndim
        return self

    def transform(self, y):
        """Makes 1D y 2D.

        Args:
            y : np.ndarray
                Target y to be transformed.

        Returns:
            np.ndarray
                A numpy array, of dimension at least 2.
        """
        if y.ndim == 1:
            return y.reshape(-1, 1)
        return y

    def inverse_transform(self, y):
        """Revert the transformation of transform.

        Args:
            y: np.ndarray
                Transformed numpy array.

        Returns:
            np.ndarray
                If the transformer was fit to a 1D numpy array,
                and a 2D numpy array with a singleton second dimension
                is passed, it will be squeezed back to 1D. Otherwise, it
                will eb left untouched.
        """
        from sklearn.utils.validation import check_is_fitted

        check_is_fitted(self)
        if self.ndim_ == 1 and y.ndim == 2:
            return np.squeeze(y, axis=1)
        return y



from keras.src import backend
from keras.src.api_export import keras_export
from keras.src.layers.preprocessing.image_preprocessing.base_image_preprocessing_layer import (  # noqa: E501
    BaseImagePreprocessingLayer,
)
from keras.src.layers.preprocessing.image_preprocessing.base_image_preprocessing_layer import (  # noqa: E501
    base_image_preprocessing_color_example,
)


@keras_export("keras.layers.RandomGrayscale")
class RandomGrayscale(BaseImagePreprocessingLayer):
    """Preprocessing layer for random conversion of RGB images to grayscale.

    This layer randomly converts input images to grayscale with a specified
    factor. When applied, it maintains the original number of channels
    but sets all channels to the same grayscale value. This can be useful
    for data augmentation and training models to be robust to color
    variations.

    The conversion preserves the perceived luminance of the original color
    image using standard RGB to grayscale conversion coefficients. Images
    that are not selected for conversion remain unchanged.

    **Note:** This layer is safe to use inside a `tf.data` or `grain` pipeline
    (independently of which backend you're using).

    Args:
        factor: Float between 0 and 1, specifying the factor of
            converting each image to grayscale. Defaults to 0.5. A value of
            1.0 means all images will be converted, while 0.0 means no images
            will be converted.
        data_format: String, one of `"channels_last"` (default) or
            `"channels_first"`. The ordering of the dimensions in the inputs.
            `"channels_last"` corresponds to inputs with shape
            `(batch, height, width, channels)` while `"channels_first"`
            corresponds to inputs with shape
            `(batch, channels, height, width)`.

    Input shape:
        3D (unbatched) or 4D (batched) tensor with shape:
        `(..., height, width, channels)`, in `"channels_last"` format,
        or `(..., channels, height, width)`, in `"channels_first"` format.

    Output shape:
        Same as input shape. The output maintains the same number of channels
        as the input, even for grayscale-converted images where all channels
        will have the same value.

    Example:

    {{base_image_preprocessing_color_example}}
    """

    def __init__(self, factor=0.5, data_format=None, seed=None, **kwargs):
        super().__init__(**kwargs)
        if factor < 0 or factor > 1:
            raise ValueError(
                f"`factor` should be between 0 and 1. Received: factor={factor}"
            )
        self.factor = factor
        self.data_format = backend.standardize_data_format(data_format)
        self.seed = seed
        self.generator = self.backend.random.SeedGenerator(seed)

    def get_random_transformation(self, images, training=True, seed=None):
        if seed is None:
            seed = self._get_seed_generator(self.backend._backend)
        # Base case: Unbatched data
        batch_size = 1
        if len(images.shape) == 4:
            # This is a batch of images (4D input)
            batch_size = self.backend.core.shape(images)[0]

        random_values = self.backend.random.uniform(
            shape=(batch_size,),
            minval=0,
            maxval=1,
            seed=seed,
        )
        should_apply = self.backend.numpy.expand_dims(
            random_values < self.factor, axis=[1, 2, 3]
        )
        return should_apply

    def transform_images(self, images, transformation, training=True):
        if training:
            should_apply = (
                transformation
                if transformation is not None
                else self.get_random_transformation(images)
            )

            grayscale_images = self.backend.image.rgb_to_grayscale(
                images, data_format=self.data_format
            )
            return self.backend.numpy.where(
                should_apply, grayscale_images, images
            )
        return images

    def compute_output_shape(self, input_shape):
        return input_shape

    def compute_output_spec(self, inputs, **kwargs):
        return backend.KerasTensor(
            inputs.shape, dtype=inputs.dtype, sparse=inputs.sparse
        )

    def transform_bounding_boxes(self, bounding_boxes, **kwargs):
        return bounding_boxes

    def transform_labels(self, labels, transformations=None, **kwargs):
        return labels

    def transform_segmentation_masks(
        self, segmentation_masks, transformations=None, **kwargs
    ):
        return segmentation_masks

    def get_config(self):
        config = super().get_config()
        config.update({"factor": self.factor})
        return config


if RandomGrayscale.__doc__ is not None:
    RandomGrayscale.__doc__ = RandomGrayscale.__doc__.replace(
        "{{base_image_preprocessing_color_example}}",
        base_image_preprocessing_color_example.replace(
            "{LayerName}", "RandomGrayscale"
        ),
    )



import math

import ml_dtypes

from keras.src import activations
from keras.src import constraints
from keras.src import initializers
from keras.src import ops
from keras.src import quantizers
from keras.src import regularizers
from keras.src.api_export import keras_export
from keras.src.layers.input_spec import InputSpec
from keras.src.layers.layer import Layer
from keras.src.quantizers.quantization_config import QuantizationConfig
from keras.src.quantizers.quantization_config import get_block_size_for_layer
from keras.src.quantizers.quantizers import dequantize_with_sz_map
from keras.src.saving import serialization_lib


@keras_export("keras.layers.Dense")
class Dense(Layer):
    """Just your regular densely-connected NN layer.

    `Dense` implements the operation:
    `output = activation(dot(input, kernel) + bias)`
    where `activation` is the element-wise activation function
    passed as the `activation` argument, `kernel` is a weights matrix
    created by the layer, and `bias` is a bias vector created by the layer
    (only applicable if `use_bias` is `True`). When this layer is
    followed by a `BatchNormalization` layer, it is recommended to set
    `use_bias=False` as `BatchNormalization` has its own bias term.

    Note: If the input to the layer has a rank greater than 2, `Dense`
    computes the dot product between the `inputs` and the `kernel` along the
    last axis of the `inputs` and axis 0 of the `kernel` (using `tf.tensordot`).
    For example, if input has dimensions `(batch_size, d0, d1)`, then we create
    a `kernel` with shape `(d1, units)`, and the `kernel` operates along axis 2
    of the `input`, on every sub-tensor of shape `(1, 1, d1)` (there are
    `batch_size * d0` such sub-tensors). The output in this case will have
    shape `(batch_size, d0, units)`.

    Args:
        units: Positive integer, dimensionality of the output space.
        activation: Activation function to use.
            If you don't specify anything, no activation is applied
            (ie. "linear" activation: `a(x) = x`).
        use_bias: Boolean, whether the layer uses a bias vector.
        kernel_initializer: Initializer for the `kernel` weights matrix.
        bias_initializer: Initializer for the bias vector.
        kernel_regularizer: Regularizer function applied to
            the `kernel` weights matrix.
        bias_regularizer: Regularizer function applied to the bias vector.
        activity_regularizer: Regularizer function applied to
            the output of the layer (its "activation").
        kernel_constraint: Constraint function applied to
            the `kernel` weights matrix.
        bias_constraint: Constraint function applied to the bias vector.
        lora_rank: Optional integer. If set, the layer's forward pass
            will implement LoRA (Low-Rank Adaptation)
            with the provided rank. LoRA sets the layer's kernel
            to non-trainable and replaces it with a delta over the
            original kernel, obtained via multiplying two lower-rank
            trainable matrices. This can be useful to reduce the
            computation cost of fine-tuning large dense layers.
            You can also enable LoRA on an existing
            `Dense` layer by calling `layer.enable_lora(rank)`.
        lora_alpha: Optional integer. If set, this parameter scales the
            low-rank adaptation delta (computed as the product of two lower-rank
            trainable matrices) during the forward pass. The delta is scaled by
            `lora_alpha / lora_rank`, allowing you to fine-tune the strength of
            the LoRA adjustment independently of `lora_rank`.

    Input shape:
        N-D tensor with shape: `(batch_size, ..., input_dim)`.
        The most common situation would be
        a 2D input with shape `(batch_size, input_dim)`.

    Output shape:
        N-D tensor with shape: `(batch_size, ..., units)`.
        For instance, for a 2D input with shape `(batch_size, input_dim)`,
        the output would have shape `(batch_size, units)`.
    """

    def __init__(
        self,
        units,
        activation=None,
        use_bias=True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        lora_rank=None,
        lora_alpha=None,
        quantization_config=None,
        **kwargs,
    ):
        if not isinstance(units, int) or units <= 0:
            raise ValueError(
                "Received an invalid value for `units`, expected a positive "
                f"integer. Received: units={units}"
            )

        super().__init__(activity_regularizer=activity_regularizer, **kwargs)
        self.units = units
        self.activation = activations.get(activation)
        self.use_bias = use_bias
        self.kernel_initializer = initializers.get(kernel_initializer)
        self.bias_initializer = initializers.get(bias_initializer)
        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        self.bias_regularizer = regularizers.get(bias_regularizer)
        self.kernel_constraint = constraints.get(kernel_constraint)
        self.bias_constraint = constraints.get(bias_constraint)
        self.lora_rank = lora_rank
        self.lora_alpha = lora_alpha if lora_alpha is not None else lora_rank
        self.lora_enabled = False
        self.quantization_config = quantization_config
        self.input_spec = InputSpec(min_ndim=2)
        self.supports_masking = True

    def build(self, input_shape):
        kernel_shape = (input_shape[-1], self.units)
        if self.quantization_mode:
            self.quantized_build(
                kernel_shape,
                mode=self.quantization_mode,
                config=self.quantization_config,
            )
        if self.quantization_mode not in ("int8", "int4", "gptq", "awq"):
            # If the layer is quantized to int8 or int4, `self._kernel` will be
            # added in `self._int8_build` or `_int4_build`. Therefore, we skip
            # it here.
            self._kernel = self.add_weight(
                name="kernel",
                shape=kernel_shape,
                initializer=self.kernel_initializer,
                regularizer=self.kernel_regularizer,
                constraint=self.kernel_constraint,
            )
        if self.use_bias:
            self.bias = self.add_weight(
                name="bias",
                shape=(self.units,),
                initializer=self.bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
            )
        else:
            self.bias = None
        self.input_spec = InputSpec(min_ndim=2, axes={-1: input_shape[-1]})
        self.built = True
        if self.lora_rank:
            self.enable_lora(self.lora_rank)

    @property
    def kernel(self):
        from keras.src.quantizers import gptq_core

        if not self.built:
            raise AttributeError(
                "You must build the layer before accessing `kernel`."
            )

        mode = self.quantization_mode
        is_gptq = mode == "gptq"
        is_awq = mode == "awq"
        is_int4 = mode == "int4"
        gptq_calibrated = bool(getattr(self, "is_gptq_calibrated", False))
        awq_calibrated = bool(getattr(self, "is_awq_calibrated", False))
        gptq_bits = (
            gptq_core.get_weight_bits_for_layer(self, None) if is_gptq else None
        )

        # Decide the source tensor first (packed vs already-quantized vs plain
        # kernel)
        if is_gptq and gptq_calibrated and gptq_bits != 4:
            # calibrated GPTQ, not 4-bit, no unpacking needed
            kernel = self.quantized_kernel
        else:
            # Start with the stored kernel
            kernel = getattr(self, "_kernel", None)

            # Handle int4 unpacking cases in one place
            if is_int4:
                # unpack [in, ceil(out/2)] to [in, out]
                kernel = quantizers.unpack_int4(
                    kernel, self._orig_output_dim, axis=-1
                )
            elif is_gptq and gptq_calibrated and gptq_bits == 4:
                kernel = quantizers.unpack_int4(
                    self.quantized_kernel,
                    orig_len=self.units,
                    axis=0,
                    dtype="uint8",
                )
            elif is_awq and awq_calibrated:
                # AWQ always uses 4-bit quantization
                kernel = quantizers.unpack_int4(
                    self.quantized_kernel,
                    orig_len=self.units,
                    axis=0,
                    dtype="uint8",
                )

        # Apply LoRA once at the end.
        if self.lora_enabled:
            kernel = kernel + (self.lora_alpha / self.lora_rank) * ops.matmul(
                self.lora_kernel_a, self.lora_kernel_b
            )

        return kernel

    def call(self, inputs, training=None):
        x = ops.matmul(inputs, self.kernel)
        if self.bias is not None:
            x = ops.add(x, self.bias)
        if self.activation is not None:
            x = self.activation(x)
        return x

    def compute_output_shape(self, input_shape):
        output_shape = list(input_shape)
        output_shape[-1] = self.units
        return tuple(output_shape)

    def enable_lora(
        self,
        rank,
        lora_alpha=None,
        a_initializer="he_uniform",
        b_initializer="zeros",
    ):
        if self.kernel_constraint:
            raise ValueError(
                "Lora is incompatible with kernel constraints. "
                "In order to enable lora on this layer, remove the "
                "`kernel_constraint` argument."
            )
        if not self.built:
            raise ValueError(
                "Cannot enable lora on a layer that isn't yet built."
            )
        if self.lora_enabled:
            raise ValueError(
                "lora is already enabled. This can only be done once per layer."
            )
        if self.quantization_mode == "gptq":
            raise NotImplementedError(
                "lora is not currently supported with GPTQ quantization."
            )
        self._tracker.unlock()
        # Determine the correct input dimension for the LoRA A matrix. When
        # the layer has been int4-quantized, `self._kernel` stores a *packed*
        # representation whose first dimension is `ceil(input_dim/2)`. We
        # saved the true, *unpacked* input dimension in `self._orig_input_dim`
        # during quantization. Use it if available; otherwise fall back to the
        # first dimension of `self.kernel`.
        if self.quantization_mode == "int4" and hasattr(
            self, "_orig_input_dim"
        ):
            input_dim_for_lora = self._orig_input_dim
        else:
            input_dim_for_lora = self.kernel.shape[0]

        self.lora_kernel_a = self.add_weight(
            name="lora_kernel_a",
            shape=(input_dim_for_lora, rank),
            initializer=initializers.get(a_initializer),
            regularizer=self.kernel_regularizer,
        )
        self.lora_kernel_b = self.add_weight(
            name="lora_kernel_b",
            shape=(rank, self.kernel.shape[1]),
            initializer=initializers.get(b_initializer),
            regularizer=self.kernel_regularizer,
        )
        self._kernel.trainable = False
        self._tracker.lock()
        self.lora_enabled = True
        self.lora_rank = rank
        self.lora_alpha = lora_alpha if lora_alpha is not None else rank

    def save_own_variables(self, store):
        # Do nothing if the layer isn't yet built
        if not self.built:
            return
        mode = self.quantization_mode
        if mode not in self.variable_serialization_spec:
            raise self._quantization_mode_error(mode)

        # Kernel plus optional merged LoRA-aware scale/zero (returns
        # (kernel, None, None) for None/gptq/awq)
        kernel_value, merged_kernel_scale, merged_kernel_zero = (
            self._get_kernel_with_merged_lora()
        )
        idx = 0
        for name in self.variable_serialization_spec[mode]:
            if name == "kernel":
                store[str(idx)] = kernel_value
            elif name == "bias" and self.bias is None:
                continue
            elif name == "kernel_zero":
                if merged_kernel_zero is None:
                    # kernel_zero only exists for sub-channel int4 quantization
                    continue
                store[str(idx)] = merged_kernel_zero
            elif name == "g_idx":
                if not hasattr(self, "g_idx"):
                    # g_idx only exists for sub-channel int4 quantization
                    continue
                store[str(idx)] = self.g_idx
            elif name == "kernel_scale" and mode in ("int4", "int8"):
                # For int4/int8, the merged LoRA scale (if any) comes from
                # `_get_kernel_with_merged_lora()`
                store[str(idx)] = merged_kernel_scale
            else:
                store[str(idx)] = getattr(self, name)
            idx += 1

    def load_own_variables(self, store):
        if not self.lora_enabled:
            self._check_load_own_variables(store)
        # Do nothing if the layer isn't yet built
        if not self.built:
            return
        mode = self.quantization_mode
        if mode not in self.variable_serialization_spec:
            raise self._quantization_mode_error(mode)

        # A saved GPTQ/AWQ quantized model will always be calibrated.
        self.is_gptq_calibrated = mode == "gptq"
        self.is_awq_calibrated = mode == "awq"

        idx = 0
        for name in self.variable_serialization_spec[mode]:
            if name == "kernel":
                self._kernel.assign(store[str(idx)])
            elif name == "bias" and self.bias is None:
                continue
            elif name == "kernel_zero" and not hasattr(self, "kernel_zero"):
                # kernel_zero only exists for sub-channel int4 quantization
                continue
            elif name == "g_idx" and not hasattr(self, "g_idx"):
                # g_idx only exists for sub-channel int4 quantization
                continue
            else:
                getattr(self, name).assign(store[str(idx)])
            idx += 1
        if self.lora_enabled:
            self.lora_kernel_a.assign(ops.zeros(self.lora_kernel_a.shape))
            self.lora_kernel_b.assign(ops.zeros(self.lora_kernel_b.shape))

    def get_config(self):
        base_config = super().get_config()
        config = {
            "units": self.units,
            "activation": activations.serialize(self.activation),
            "use_bias": self.use_bias,
            "kernel_initializer": initializers.serialize(
                self.kernel_initializer
            ),
            "bias_initializer": initializers.serialize(self.bias_initializer),
            "kernel_regularizer": regularizers.serialize(
                self.kernel_regularizer
            ),
            "bias_regularizer": regularizers.serialize(self.bias_regularizer),
            "kernel_constraint": constraints.serialize(self.kernel_constraint),
            "bias_constraint": constraints.serialize(self.bias_constraint),
            "quantization_config": serialization_lib.serialize_keras_object(
                self.quantization_config
            ),
        }
        if self.lora_rank:
            config["lora_rank"] = self.lora_rank
            config["lora_alpha"] = self.lora_alpha
        return {**base_config, **config}

    @classmethod
    def from_config(cls, config):
        config = config.copy()
        config["quantization_config"] = (
            serialization_lib.deserialize_keras_object(
                config.get("quantization_config", None)
            )
        )
        return super().from_config(config)

    @property
    def variable_serialization_spec(self):
        """Returns a dict mapping quantization modes to variable names in order.

        This spec is used by `save_own_variables` and `load_own_variables` to
        determine the correct ordering of variables during serialization for
        each quantization mode. `None` means no quantization.
        """
        return {
            None: [
                "kernel",
                "bias",
            ],
            "int8": [
                "kernel",
                "bias",
                "kernel_scale",
            ],
            "int4": [
                "kernel",
                "bias",
                "kernel_scale",
                "kernel_zero",
                "g_idx",
            ],
            "float8": [
                "kernel",
                "bias",
                "inputs_scale",
                "inputs_amax_history",
                "kernel_scale",
                "kernel_amax_history",
                "outputs_grad_scale",
                "outputs_grad_amax_history",
            ],
            "gptq": [
                "bias",
                "quantized_kernel",
                "kernel_scale",
                "kernel_zero",
                "g_idx",
            ],
            "awq": [
                "bias",
                "quantized_kernel",
                "kernel_scale",
                "kernel_zero",
                "awq_scales",
                "g_idx",
            ],
        }

    def quantized_build(self, kernel_shape, mode, config=None):
        if mode == "int8":
            self._int8_build(kernel_shape, config)
        elif mode == "int4":
            self._int4_build(kernel_shape, config)
        elif mode == "float8":
            self._float8_build()
        elif mode == "gptq":
            self._gptq_build(kernel_shape, config)
        elif mode == "awq":
            self._awq_build(kernel_shape, config)
        else:
            raise self._quantization_mode_error(mode)
        self._is_quantized = True

    def _int8_build(self, kernel_shape, config=None):
        self.inputs_quantizer = (
            QuantizationConfig.activation_quantizer_or_default(
                config, quantizers.AbsMaxQuantizer()
            )
        )

        self._kernel = self.add_weight(
            name="kernel",
            shape=kernel_shape,
            initializer="zeros",
            dtype="int8",
            trainable=False,
        )
        self.kernel_scale = self.add_weight(
            name="kernel_scale",
            shape=(self.units,),
            initializer="ones",
            trainable=False,
        )

    def _gptq_build(self, kernel_shape, config):
        from keras.src.quantizers import gptq_core

        # Ensures the forward pass uses the original high-precision kernel
        # until calibration has been performed.
        self.is_gptq_calibrated = False
        self.kernel_shape = kernel_shape

        weight_bits = gptq_core.get_weight_bits_for_layer(self, config)
        # For 4-bit weights, we pack two values per byte.
        units = (
            (kernel_shape[1] + 1) // 2 if weight_bits == 4 else kernel_shape[1]
        )

        self.quantized_kernel = self.add_weight(
            name="kernel",
            shape=(units, kernel_shape[0]),
            initializer="zeros",
            dtype="uint8",
            trainable=False,
        )

        group_size = gptq_core.get_group_size_for_layer(self, config)
        n_groups = (
            1
            if group_size == -1
            else math.ceil(self.kernel_shape[0] / group_size)
        )
        self.kernel_scale = self.add_weight(
            name="kernel_scale",
            shape=(self.units, n_groups),
            initializer="ones",
            trainable=False,
        )
        self.kernel_zero = self.add_weight(
            name="kernel_zero",
            shape=(self.units, n_groups),
            initializer="zeros",
            dtype="uint8",
            trainable=False,
        )
        self.g_idx = self.add_weight(
            name="g_idx",
            shape=(self.kernel_shape[0],),
            initializer="zeros",
            dtype="float32",
            trainable=False,
        )

    def _gptq_call(self, inputs, training=False):
        from keras.src.quantizers import gptq_core

        if not self.is_gptq_calibrated:
            W = self._kernel
        else:
            should_unpack = (
                gptq_core.get_weight_bits_for_layer(self, config=None) == 4
            )
            W = (
                quantizers.unpack_int4(
                    self.quantized_kernel,
                    orig_len=self.units,
                    axis=0,
                    dtype="uint8",
                )
                if should_unpack
                else self.quantized_kernel
            )
            W = ops.transpose(
                dequantize_with_sz_map(
                    W,
                    self.kernel_scale,
                    self.kernel_zero,
                    self.g_idx,
                )
            )

        y = ops.matmul(inputs, W)
        if self.bias is not None:
            y = ops.add(y, self.bias)
        if self.activation is not None:
            y = self.activation(y)
        return y

    def _awq_build(self, kernel_shape, config):
        """Build variables for AWQ quantization.

        AWQ uses 4-bit quantization with per-channel AWQ scales that protect
        salient weights based on activation magnitudes.
        """
        from keras.src.quantizers import awq_core

        # Ensures the forward pass uses the original high-precision kernel
        # until calibration has been performed.
        self.is_awq_calibrated = False
        self.kernel_shape = kernel_shape

        # For 4-bit weights, we pack two values per byte.
        units = (kernel_shape[1] + 1) // 2

        self.quantized_kernel = self.add_weight(
            name="kernel",
            shape=(units, kernel_shape[0]),
            initializer="zeros",
            dtype="uint8",
            trainable=False,
        )

        group_size = awq_core.get_group_size_for_layer(self, config)
        num_groups = (
            1 if group_size == -1 else math.ceil(kernel_shape[0] / group_size)
        )
        self.kernel_scale = self.add_weight(
            name="kernel_scale",
            shape=(self.units, num_groups),
            initializer="ones",
            trainable=False,
        )
        self.kernel_zero = self.add_weight(
            name="kernel_zero",
            shape=(self.units, num_groups),
            initializer="zeros",
            dtype="uint8",
            trainable=False,
        )

        # Per-channel AWQ scales from activation magnitudes
        self.awq_scales = self.add_weight(
            name="awq_scales",
            shape=(kernel_shape[0],),
            initializer="ones",
            trainable=False,
        )
        self.g_idx = self.add_weight(
            name="g_idx",
            shape=(kernel_shape[0],),
            initializer="zeros",
            dtype="float32",
            trainable=False,
        )

    def _awq_call(self, inputs, training=False):
        """Forward pass for AWQ quantized layer."""
        if not self.is_awq_calibrated:
            W = self._kernel
        else:
            # Unpack 4-bit weights
            W = quantizers.unpack_int4(
                self.quantized_kernel,
                orig_len=self.units,
                axis=0,
                dtype="uint8",
            )
            # Dequantize using scale/zero maps
            W = ops.transpose(
                dequantize_with_sz_map(
                    W,
                    self.kernel_scale,
                    self.kernel_zero,
                    self.g_idx,
                )
            )
            # Apply AWQ scales by dividing to restore original magnitude
            # (We multiplied by scales before quantization, so divide to undo)
            # awq_scales has shape [input_dim], W has shape [input_dim, units]
            # Expand dims for proper broadcasting.
            W = ops.divide(W, ops.expand_dims(self.awq_scales, -1))

        y = ops.matmul(inputs, W)
        if self.bias is not None:
            y = ops.add(y, self.bias)
        if self.activation is not None:
            y = self.activation(y)
        return y

    def _int4_build(self, kernel_shape, config=None):
        """Build variables for int4 quantization.

        The kernel is packed along the last axis,
        resulting in shape `(input_dim, ceil(units/2))`.

        Args:
            kernel_shape: The original float32 kernel shape
                `(input_dim, units)`.
            config: Optional quantization config specifying block_size.
        """
        self.inputs_quantizer = (
            QuantizationConfig.activation_quantizer_or_default(config, None)
        )
        input_dim, output_dim = kernel_shape

        # kernel is packed along last axis (output dimension)
        # Stored shape: [input_dim, ceil(output_dim/2)]
        packed_cols = (output_dim + 1) // 2

        self._kernel = self.add_weight(
            name="kernel",
            shape=(input_dim, packed_cols),
            initializer="zeros",
            dtype="int8",
            trainable=False,
        )

        block_size = get_block_size_for_layer(self, config)
        self._int4_block_size = block_size

        if block_size is None or block_size == -1:
            # Per-channel: one scale per output unit
            scale_shape = (self.units,)
        else:
            # Sub-channel: [n_groups, out_features]
            n_groups = math.ceil(input_dim / block_size)
            scale_shape = (n_groups, self.units)

        self.kernel_scale = self.add_weight(
            name="kernel_scale",
            shape=scale_shape,
            initializer="ones",
            trainable=False,
        )

        # Sub-channel quantization uses asymmetric quantization
        if block_size is not None and block_size > 0:

            def idx_initializer(shape, dtype):
                return ops.floor_divide(
                    ops.arange(input_dim, dtype=dtype), block_size
                )

            self.kernel_zero = self.add_weight(
                name="kernel_zero",
                shape=scale_shape,
                initializer="zeros",
                dtype="int8",
                trainable=False,
            )
            self.g_idx = self.add_weight(
                name="g_idx",
                shape=(input_dim,),
                initializer=idx_initializer,
                dtype="float32",
                trainable=False,
            )

        # Record dimensions for unpacking and reshaping at runtime.
        self._orig_input_dim = input_dim
        self._orig_output_dim = output_dim

    def _float8_build(self):
        from keras.src.dtype_policies import QuantizedFloat8DTypePolicy

        # If `self.dtype_policy` is not QuantizedFloat8DTypePolicy, then set
        # `amax_history_length` to its default value.
        amax_history_length = getattr(
            self.dtype_policy,
            "amax_history_length",
            QuantizedFloat8DTypePolicy.default_amax_history_length,
        )
        # We set `trainable=True` because we will use the gradients to overwrite
        # these variables
        scale_kwargs = {
            "shape": (),
            "initializer": "ones",
            "dtype": "float32",  # Always be float32
            "trainable": True,
            "autocast": False,
            "overwrite_with_gradient": True,
        }
        amax_history_kwargs = {
            "shape": (amax_history_length,),
            "initializer": "zeros",
            "dtype": "float32",  # Always be float32
            "trainable": True,
            "autocast": False,
            "overwrite_with_gradient": True,
        }
        self.inputs_scale = self.add_weight(name="inputs_scale", **scale_kwargs)
        self.inputs_amax_history = self.add_weight(
            name="inputs_amax_history", **amax_history_kwargs
        )
        self.kernel_scale = self.add_weight(name="kernel_scale", **scale_kwargs)
        self.kernel_amax_history = self.add_weight(
            name="kernel_amax_history", **amax_history_kwargs
        )
        self.outputs_grad_scale = self.add_weight(
            name="outputs_grad_scale", **scale_kwargs
        )
        self.outputs_grad_amax_history = self.add_weight(
            name="outputs_grad_amax_history", **amax_history_kwargs
        )

    def _int8_call(self, inputs, training=None):
        @ops.custom_gradient
        def matmul_with_inputs_gradient(inputs, kernel, kernel_scale):
            """Custom gradient function to handle the int8 quantized weights.

            Automatic differentiation will not know how to handle the int8
            quantized weights. So a custom gradient function is needed to
            handle the int8 quantized weights.

            The custom gradient function will use the dequantized kernel to
            compute the gradient.
            """

            def grad_fn(*args, upstream=None):
                if upstream is None:
                    (upstream,) = args
                float_kernel = ops.divide(
                    ops.cast(kernel, dtype=self.compute_dtype),
                    kernel_scale,
                )
                inputs_grad = ops.matmul(upstream, ops.transpose(float_kernel))
                return (inputs_grad, None, None)

            output_scale = kernel_scale
            if self.inputs_quantizer:
                inputs, inputs_scale = self.inputs_quantizer(inputs, axis=-1)
                output_scale = ops.multiply(output_scale, inputs_scale)

            x = ops.matmul(inputs, kernel)
            # De-scale outputs
            x = ops.cast(x, self.compute_dtype)
            x = ops.divide(x, output_scale)
            return x, grad_fn

        x = matmul_with_inputs_gradient(
            inputs,
            ops.convert_to_tensor(self._kernel),
            ops.convert_to_tensor(self.kernel_scale),
        )
        if self.lora_enabled:
            lora_x = ops.matmul(inputs, self.lora_kernel_a)
            lora_x = ops.matmul(lora_x, self.lora_kernel_b)
            x = ops.add(x, (self.lora_alpha / self.lora_rank) * lora_x)
        if self.bias is not None:
            x = ops.add(x, self.bias)
        if self.activation is not None:
            x = self.activation(x)
        return x

    def _int4_call(self, inputs, training=None):
        """Forward pass for int4 quantized Dense layer.

        Uses custom gradients to handle quantized weights since autodiff
        cannot differentiate through int4 operations.
        """
        block_size = getattr(self, "_int4_block_size", None)

        if block_size is None or block_size == -1:
            # Per-channel: symmetric quantization (no zero point needed)
            @ops.custom_gradient
            def matmul_per_channel_with_inputs_gradient(
                inputs, kernel, kernel_scale
            ):
                """Per-channel int4 forward pass with custom gradient."""
                # Unpack: stored as [in, ceil(out/2)], unpack along last axis
                unpacked_kernel = quantizers.unpack_int4(
                    kernel, self._orig_output_dim, axis=-1
                )

                def grad_fn(*args, upstream=None):
                    if upstream is None:
                        (upstream,) = args
                    # Per-channel: unpacked is [in, out]
                    float_kernel = ops.divide(
                        ops.cast(unpacked_kernel, dtype=self.compute_dtype),
                        kernel_scale,
                    )
                    inputs_grad = ops.matmul(
                        upstream, ops.transpose(float_kernel)
                    )
                    return (inputs_grad, None, None)

                # Forward pass: per-channel dequantization
                output_scale = kernel_scale
                if self.inputs_quantizer:
                    inputs, inputs_scale = self.inputs_quantizer(
                        inputs, axis=-1
                    )
                    output_scale = ops.multiply(output_scale, inputs_scale)

                x = ops.matmul(inputs, unpacked_kernel)
                x = ops.cast(x, self.compute_dtype)
                x = ops.divide(x, output_scale)
                return x, grad_fn

            x = matmul_per_channel_with_inputs_gradient(
                inputs,
                ops.convert_to_tensor(self._kernel),
                ops.convert_to_tensor(self.kernel_scale),
            )
        else:
            # Sub-channel: asymmetric quantization (with zero point)
            @ops.custom_gradient
            def matmul_sub_channel_with_inputs_gradient(
                inputs, kernel, kernel_scale, kernel_zero, g_idx
            ):
                """Sub-channel int4 forward pass with custom gradient."""
                # Unpack: stored as [in, ceil(out/2)], unpack along last axis
                unpacked_kernel = quantizers.unpack_int4(
                    kernel, self._orig_output_dim, axis=-1
                )

                def grad_fn(*args, upstream=None):
                    if upstream is None:
                        (upstream,) = args
                    float_kernel = dequantize_with_sz_map(
                        unpacked_kernel,
                        kernel_scale,
                        kernel_zero,
                        g_idx,
                        group_axis=0,
                    )
                    float_kernel = ops.cast(float_kernel, self.compute_dtype)
                    inputs_grad = ops.matmul(
                        upstream, ops.transpose(float_kernel)
                    )
                    return (inputs_grad, None, None, None, None)

                float_kernel = dequantize_with_sz_map(
                    unpacked_kernel,
                    kernel_scale,
                    kernel_zero,
                    g_idx,
                    group_axis=0,
                )
                float_kernel = ops.cast(float_kernel, self.compute_dtype)
                x = ops.matmul(inputs, float_kernel)
                return x, grad_fn

            x = matmul_sub_channel_with_inputs_gradient(
                inputs,
                ops.convert_to_tensor(self._kernel),
                ops.convert_to_tensor(self.kernel_scale),
                ops.convert_to_tensor(self.kernel_zero),
                ops.convert_to_tensor(self.g_idx),
            )

        if self.lora_enabled:
            lora_x = ops.matmul(inputs, self.lora_kernel_a)
            lora_x = ops.matmul(lora_x, self.lora_kernel_b)
            x = ops.add(x, (self.lora_alpha / self.lora_rank) * lora_x)

        if self.bias is not None:
            x = ops.add(x, self.bias)
        if self.activation is not None:
            x = self.activation(x)
        return x

    def _float8_call(self, inputs, training=None):
        if self.lora_enabled:
            raise NotImplementedError(
                "Currently, `_float8_call` doesn't support LoRA"
            )

        @ops.custom_gradient
        def quantized_dequantize_inputs(inputs, scale, amax_history):
            if training:
                new_scale = quantizers.compute_float8_scale(
                    ops.max(amax_history, axis=0),
                    scale,
                    ops.cast(
                        float(ml_dtypes.finfo("float8_e4m3fn").max), "float32"
                    ),
                )
                new_amax_history = quantizers.compute_float8_amax_history(
                    inputs, amax_history
                )
            else:
                new_scale = None
                new_amax_history = None
            qdq_inputs = quantizers.quantize_and_dequantize(
                inputs, scale, "float8_e4m3fn", self.compute_dtype
            )

            def grad(*args, upstream=None, variables=None):
                if upstream is None:
                    (upstream,) = args
                return upstream, new_scale, new_amax_history

            return qdq_inputs, grad

        @ops.custom_gradient
        def quantized_dequantize_outputs(outputs, scale, amax_history):
            """Quantize-dequantize the output gradient but not the output."""

            def grad(*args, upstream=None, variables=None):
                if upstream is None:
                    (upstream,) = args
                new_scale = quantizers.compute_float8_scale(
                    ops.max(amax_history, axis=0),
                    scale,
                    ops.cast(
                        float(ml_dtypes.finfo("float8_e5m2").max), "float32"
                    ),
                )
                qdq_upstream = quantizers.quantize_and_dequantize(
                    upstream, scale, "float8_e5m2", self.compute_dtype
                )
                new_amax_history = quantizers.compute_float8_amax_history(
                    upstream, amax_history
                )
                return qdq_upstream, new_scale, new_amax_history

            return outputs, grad

        x = ops.matmul(
            quantized_dequantize_inputs(
                inputs,
                ops.convert_to_tensor(self.inputs_scale),
                ops.convert_to_tensor(self.inputs_amax_history),
            ),
            quantized_dequantize_inputs(
                ops.convert_to_tensor(self._kernel),
                ops.convert_to_tensor(self.kernel_scale),
                ops.convert_to_tensor(self.kernel_amax_history),
            ),
        )
        # `quantized_dequantize_outputs` is placed immediately after
        # `ops.matmul` for the sake of pattern matching in gemm_rewrite. That
        # way, the qdq will be adjacent to the corresponding matmul_bprop in the
        # bprop.
        x = quantized_dequantize_outputs(
            x,
            ops.convert_to_tensor(self.outputs_grad_scale),
            ops.convert_to_tensor(self.outputs_grad_amax_history),
        )
        if self.bias is not None:
            # Under non-mixed precision cases, F32 bias has to be converted to
            # BF16 first to get the biasAdd fusion support. ref. PR
            # https://github.com/tensorflow/tensorflow/pull/60306
            bias = self.bias
            if self.dtype_policy.compute_dtype == "float32":
                bias_bf16 = ops.cast(bias, "bfloat16")
                bias = ops.cast(bias_bf16, bias.dtype)
            x = ops.add(x, bias)
        if self.activation is not None:
            x = self.activation(x)
        return x

    def quantize(self, mode=None, type_check=True, config=None):
        # Prevent quantization of the subclasses
        if type_check and (type(self) is not Dense):
            raise self._not_implemented_error(self.quantize)

        self.quantization_config = config

        kernel_shape = self._kernel.shape
        if mode == "int8":
            weight_quantizer = QuantizationConfig.weight_quantizer_or_default(
                self.quantization_config, quantizers.AbsMaxQuantizer(axis=0)
            )
            kernel_value, kernel_scale = weight_quantizer(
                self._kernel, to_numpy=True
            )
            kernel_scale = ops.squeeze(kernel_scale, axis=0)
            del self._kernel
            # Build variables for int8 mode
            self.quantized_build(kernel_shape, mode, self.quantization_config)
            self._kernel.assign(kernel_value)
            self.kernel_scale.assign(kernel_scale)
        elif mode == "int4":
            from keras.src.quantizers.quantization_config import (
                Int4QuantizationConfig,
            )

            block_size = None
            if isinstance(self.quantization_config, Int4QuantizationConfig):
                block_size = self.quantization_config.block_size

            if block_size is None or block_size == -1:
                # Per-channel quantization
                weight_quantizer = (
                    QuantizationConfig.weight_quantizer_or_default(
                        self.quantization_config,
                        quantizers.AbsMaxQuantizer(
                            axis=0, value_range=(-8, 7), output_dtype="int8"
                        ),
                    )
                )
                kernel_value_int4, kernel_scale = weight_quantizer(
                    self._kernel, to_numpy=True
                )
                kernel_scale = ops.squeeze(kernel_scale, axis=0)
            else:
                # Sub-channel quantization with asymmetric zero point
                # Returns kernel [in, out], scale [n_groups, out], zero
                # [n_groups, out]
                kernel_value_int4, kernel_scale, kernel_zero = (
                    quantizers.abs_max_quantize_grouped_with_zero_point(
                        self._kernel, block_size=block_size, to_numpy=True
                    )
                )

            # Pack two int4 values per int8 byte along last axis
            # Stored as [in, ceil(out/2)]
            packed_kernel_value, _, _ = quantizers.pack_int4(
                kernel_value_int4, axis=-1
            )
            del self._kernel
            self.quantized_build(kernel_shape, mode, self.quantization_config)
            self._kernel.assign(packed_kernel_value)
            self.kernel_scale.assign(kernel_scale)
            if block_size is not None and block_size > 0:
                self.kernel_zero.assign(kernel_zero)
        elif mode == "gptq":
            self.quantized_build(kernel_shape, mode, self.quantization_config)
        elif mode == "awq":
            self.quantized_build(kernel_shape, mode, self.quantization_config)
        elif mode == "float8":
            self.quantized_build(kernel_shape, mode)
        else:
            raise self._quantization_mode_error(mode)

        # Set new dtype policy only for modes that already have a policy.
        if self.dtype_policy.quantization_mode is None:
            from keras.src import dtype_policies  # local import to avoid cycle

            policy_name = mode
            if mode in ("gptq", "awq"):
                policy_name = self.quantization_config.dtype_policy_string()
            elif mode == "int4":
                # Include block_size in policy name for sub-channel quantization
                block_size = get_block_size_for_layer(self, config)
                # Use -1 for per-channel, otherwise use block_size
                block_size_value = -1 if block_size is None else block_size
                policy_name = f"int4/{block_size_value}"
            policy = dtype_policies.get(
                f"{policy_name}_from_{self.dtype_policy.name}"
            )
            self.dtype_policy = policy

    def _get_kernel_with_merged_lora(self):
        """Returns the kernel with LoRA matrices merged, for serialization.

        This method is called by `save_own_variables` to produce a single
        kernel tensor that includes the adaptations from LoRA. This is useful
        for deploying the model or for continuing training after permanently
        applying the LoRA update.

        If the layer is quantized (`int8` or `int4`), the process is:
        1. Dequantize the base kernel to float.
        2. Compute the LoRA delta (`lora_kernel_a @ lora_kernel_b`) and add
            it to the dequantized kernel.
        3. Re-quantize the merged result back to the original quantized
            type (`int8` or packed `int4`), calculating a new scale factor.

        If the layer is not quantized, this method returns the result of the
        `kernel` property (which computes the merge in floating-point) and a
        scale of `None`.

        If LoRA is not enabled, it returns the original kernel and scale
        without modification.

        Returns:
            A tuple `(kernel_value, kernel_scale, kernel_zero)`:
                `kernel_value`: The merged kernel. A quantized tensor if
                    quantization is active, otherwise a high precision tensor.
                `kernel_scale`: The quantization scale for the merged kernel.
                    This is `None` if the layer is not quantized.
                `kernel_zero`: The zero point for sub-channel int4 quantization.
                    This is `None` for per-channel or non-int4 modes.
        """
        if self.dtype_policy.quantization_mode in (None, "gptq", "awq"):
            return self.kernel, None, None

        kernel_value = self._kernel
        kernel_scale = self.kernel_scale
        kernel_zero = getattr(self, "kernel_zero", None)

        if not self.lora_enabled:
            return kernel_value, kernel_scale, kernel_zero

        # Dequantize, Merge, and Re-quantize
        block_size = getattr(self, "_int4_block_size", None)

        # Step 1: Dequantize kernel to float
        if self.quantization_mode == "int4":
            # Unpack along last axis ([in, out])
            unpacked_kernel = quantizers.unpack_int4(
                kernel_value, self._orig_output_dim, axis=-1
            )
            if block_size is None or block_size == -1:
                # Per-channel: kernel [in, out], scale [out]
                float_kernel = ops.divide(
                    ops.cast(unpacked_kernel, self.compute_dtype),
                    kernel_scale,
                )
            else:
                # Sub-channel: scale/zero are [n_groups, out]
                float_kernel = dequantize_with_sz_map(
                    unpacked_kernel,
                    kernel_scale,
                    self.kernel_zero,
                    self.g_idx,
                    group_axis=0,
                )
                float_kernel = ops.cast(float_kernel, self.compute_dtype)
            quant_range = (-8, 7)
        elif self.quantization_mode == "int8":
            float_kernel = ops.divide(
                ops.cast(kernel_value, self.compute_dtype), kernel_scale
            )
            quant_range = (-127, 127)
        else:
            raise ValueError(
                f"Unsupported quantization mode: {self.quantization_mode}"
            )

        # Step 2: Merge LoRA weights in float domain
        lora_delta = (self.lora_alpha / self.lora_rank) * ops.matmul(
            self.lora_kernel_a, self.lora_kernel_b
        )
        merged_float_kernel = ops.add(float_kernel, lora_delta)

        # Step 3: Re-quantize the merged kernel
        if (
            self.quantization_mode == "int4"
            and block_size is not None
            and block_size != -1
        ):
            # Sub-channel: returns kernel [in, out], scale [n_groups, out]
            requantized_kernel, kernel_scale, kernel_zero = (
                quantizers.abs_max_quantize_grouped_with_zero_point(
                    merged_float_kernel, block_size=block_size, to_numpy=True
                )
            )
        elif self.quantization_mode == "int4":
            # Per-channel: quantize along input axis (axis=0)
            requantized_kernel, kernel_scale = quantizers.abs_max_quantize(
                merged_float_kernel,
                axis=0,
                value_range=quant_range,
                dtype="int8",
                to_numpy=True,
            )
            kernel_scale = ops.squeeze(kernel_scale, axis=0)
            kernel_zero = None
        else:
            requantized_kernel, kernel_scale = quantizers.abs_max_quantize(
                merged_float_kernel,
                axis=0,
                value_range=quant_range,
                dtype="int8",
                to_numpy=True,
            )
            kernel_scale = ops.squeeze(kernel_scale, axis=0)
            kernel_zero = None

        if self.quantization_mode == "int4":
            # Pack along last axis
            kernel_value, _, _ = quantizers.pack_int4(
                requantized_kernel, axis=-1
            )
        else:
            kernel_value = requantized_kernel
        return kernel_value, kernel_scale, kernel_zero



import numpy as np

from keras.src.api_export import keras_export
from keras.src.callbacks.callback import Callback
from keras.src.utils import io_utils


@keras_export("keras.callbacks.TerminateOnNaN")
class TerminateOnNaN(Callback):
    """Callback that terminates training when a NaN loss is encountered.

    This callback monitors the loss value during training
    and terminates training when a NaN or Inf loss is detected.
    By default, training is stopped gracefully
    by setting `model.stop_training = True`, which triggers all callback cleanup
    methods including `on_train_end()`.

    Alternatively, you can use `raise_error=True` to immediately raise a
    RuntimeError when NaN/Inf is detected. This raise_error termination
    prevents `on_train_end()` from being called on other callbacks, which
    is useful for preserving backup states or preventing unintended cleanup
    when training fails.

    Args:
        raise_error: Boolean, default False. If False, uses graceful stop via
            `model.stop_training = True`. If True, immediately raises
            RuntimeError on NaN/Inf loss, bypassing callback cleanup methods.

    Example:

    ```
    # Graceful termination (default)
    callback = keras.callbacks.TerminateOnNaN()
    model.fit(x, y, callbacks=[callback])

    # raise_error termination (strict failure)
    callback = keras.callbacks.TerminateOnNaN(raise_error=True)
    model.fit(x, y, callbacks=[callback])
    ```
    """

    def __init__(self, raise_error: bool = False):
        super().__init__()
        self.raise_error = raise_error

    def on_batch_end(self, batch, logs=None):
        """Check for NaN/Inf loss at the end of each batch.

        Args:
            batch: Integer, index of batch within the current epoch.
            logs: Dict, contains the return value of `model.train_step()`.

        Raises:
            RuntimeError: If loss is NaN/Inf and raise_error=True.
        """
        logs = logs or {}
        loss = logs.get("loss")
        if loss is not None:
            if np.isnan(loss) or np.isinf(loss):
                if self.raise_error:
                    raise RuntimeError(
                        f"NaN or Inf loss encountered at batch {batch}. "
                        f"Loss value: {loss}. Terminating training immediately."
                    )
                else:
                    io_utils.print_msg(
                        f"Batch {batch}: Invalid loss, terminating training"
                    )
                    self.model.stop_training = True



import re

from keras.src import ops
from keras.src.api_export import keras_export
from keras.src.optimizers import optimizer


@keras_export(["keras.optimizers.Muon"])
class Muon(optimizer.Optimizer):
    """Optimizer that implements the Muon algorithm.

    Note that this optimizer should not be used in the following layers:

    1. Embedding layer
    2. Final output fully connected layer
    3. Any {0,1}-D variables

    These should all be optimized using AdamW.

    The Muon optimizer can use both the Muon update step or the
    AdamW update step based on the following:

    - For any variable that isn't 2D, the AdamW step
        will be used. This is not configurable.
    - If the argument `exclude_embeddings` (defaults to `True`) is set
    to `True`, the AdamW step will be used.
    - For any variablewith a name that matches an expression
        listed in the argument `exclude_layers` (a list), the
        AdamW step will be used.
    - Any other variable uses the Muon step.

    Typically, you only need to pass the name of your densely-connected
    output layer to `exclude_layers`, e.g.
    `exclude_layers=["output_dense"]`.

    References:
        - [Original implementation](https://github.com/KellerJordan/Muon)
        - [Liu et al, 2025](https://arxiv.org/abs/2502.16982)

    Args:
        learning_rate: A float,
            `keras.optimizers.schedules.LearningRateSchedule` instance, or
            a callable that takes no arguments and returns the actual value to
            use. The learning rate. Defaults to `0.001`.
        adam_beta_1: A float value or a constant float tensor, or a callable
            that takes no arguments and returns the actual value to use.
            The exponential decay rate for the 1st moment estimates. Defaults to
            `0.9`.
        adam_beta_2: A float value or a constant float tensor, or a callable
            that takes no arguments and returns the actual value to use.
            The exponential decay rate for the 2nd moment estimates. Defaults to
            `0.999`.
        adam_weight_decay: Float. If set, weight decay is applied when using
            the Adam optimizer.
        epsilon: A small constant for numerical stability. This is
            "epsilon hat" in the Kingma and Ba paper
            (in the formula just before Section 2.1),
            not the epsilon in Algorithm 1 of the paper.
            It be used at Adamw.Defaults to `1e-7`.
        exclude_layers: List of strings, keywords of layer names to exclude.
            All layers with keywords in their path will use adamw.
        exclude_embeddings: Boolean value
            If True, embedding layers will use adamw.
        muon_a: Float, parameter a of the muon algorithm.
            It is recommended to use the default value
        muon_b: Float, parameter b of the muon algorithm.
            It is recommended to use the default value
        muon_c: Float, parameter c of the muon algorithm.
            It is recommended to use the default value
        adam_lr_ratio: Float, the ratio of the learning rate when
                using Adam to the main learning rate.
                It is recommended to set it to 1
        momentum: Float, momentum used by internal SGD.
        ns_steps: Integer, number of Newton-Schulz iterations to run.
        nesterov: Boolean, whether to use Nesterov-style momentum
        {{base_optimizer_keyword_args}}
        rms_rate: Float. A parameter from https://arxiv.org/abs/2502.16982
            that can enhance the stability of Muon, allowing it to use the
            same learning rate and weight decay as Adam. Defaults to `0.2`.
            Set to `None` to disable this feature.
    """

    def __init__(
        self,
        learning_rate=0.001,
        adam_beta_1=0.9,
        adam_beta_2=0.999,
        adam_weight_decay=0.004,
        epsilon=1e-7,
        weight_decay=0.004,
        clipnorm=None,
        clipvalue=None,
        global_clipnorm=None,
        use_ema=False,
        ema_momentum=0.99,
        ema_overwrite_frequency=None,
        loss_scale_factor=None,
        gradient_accumulation_steps=None,
        name="muon",
        exclude_layers=None,
        exclude_embeddings=True,
        muon_a=3.4445,
        muon_b=-4.7750,
        muon_c=2.0315,
        adam_lr_ratio=1,
        momentum=0.95,
        ns_steps=5,
        nesterov=True,
        rms_rate=0.2,
        **kwargs,
    ):
        super().__init__(
            learning_rate=learning_rate,
            name=name,
            weight_decay=weight_decay,
            clipnorm=clipnorm,
            clipvalue=clipvalue,
            global_clipnorm=global_clipnorm,
            use_ema=use_ema,
            ema_momentum=ema_momentum,
            ema_overwrite_frequency=ema_overwrite_frequency,
            loss_scale_factor=loss_scale_factor,
            gradient_accumulation_steps=gradient_accumulation_steps,
            **kwargs,
        )
        self.adam_beta_1 = adam_beta_1
        self.adam_beta_2 = adam_beta_2
        self.epsilon = epsilon
        self.muon_a = muon_a
        self.muon_b = muon_b
        self.muon_c = muon_c
        self.adam_lr_ratio = adam_lr_ratio
        self.momentum = momentum
        self.ns_steps = ns_steps
        self.nesterov = nesterov
        self.exclude_embeddings = exclude_embeddings
        self.exclude_layers = exclude_layers or []
        self.adam_weight_decay = adam_weight_decay
        self.rms_rate = rms_rate

    def _should_use_adamw(self, variable):
        # it works well to just flatten their last 3 dimensions.
        # any {0,1}-D parameters should all be optimized by adam
        if len(variable.shape) != 2:
            return True
        if self.exclude_embeddings and "embedding" in variable.path.lower():
            return True
        for keyword in self.exclude_layers:
            if re.search(keyword, variable.path):
                return True
        return False

    def build(self, var_list):
        """Initialize optimizer variables.

        Adam optimizer has 3 types of variables: momentums, velocities and
        velocity_hat (only set when amsgrad is applied),

        Args:
            var_list: list of model variables to build Adam variables on.
        """
        if self.built:
            return
        super().build(var_list)
        # Momentums are for both Muon and Adam
        self.momentums = [None] * len(var_list)
        # Velocities are just for Adam
        self.adam_velocities = [None] * len(var_list)

        for var in var_list:
            if not self._overwrite_variable_with_gradient(var):
                self.momentums[self._get_variable_index(var)] = (
                    self.add_variable_from_reference(
                        reference_variable=var, name="momentum"
                    )
                )
                if self._should_use_adamw(var):
                    self.adam_velocities[self._get_variable_index(var)] = (
                        self.add_variable_from_reference(
                            reference_variable=var, name="velocity"
                        )
                    )

    def update_step(self, gradient, variable, learning_rate):
        variable_index = self._get_variable_index(variable)
        m = self.momentums[variable_index]
        v = self.adam_velocities[variable_index]

        # The presence of the velocity tells us that this variable is for Adam
        if v is not None:
            # It should be noted that lr is one-tenth when using adamw.
            self._adamw_update_step(
                gradient, variable, learning_rate * self.adam_lr_ratio, m, v
            )
        else:
            self._muon_update_step(gradient, variable, learning_rate, m)

    def _muon_update_step(self, gradient, variable, lr, m):
        self.assign_add(m, ops.add(gradient, m * (self.momentum - 1)))
        if self.nesterov:
            g = ops.add(gradient, self.momentum * m)
        else:
            g = m
        update = self.zeropower_via_newtonschulz5(g, self.ns_steps)

        self.assign_sub(variable, self.lr_adjust(lr * update))

    def _adamw_update_step(self, gradient, variable, learning_rate, m, v):
        """Update step given gradient and the associated model variable."""
        lr = ops.cast(learning_rate, variable.dtype)
        gradient = ops.cast(gradient, variable.dtype)
        local_step = ops.cast(self.iterations + 1, variable.dtype)
        adam_beta_1_power = ops.power(
            ops.cast(self.adam_beta_1, variable.dtype), local_step
        )
        adam_beta_2_power = ops.power(
            ops.cast(self.adam_beta_2, variable.dtype), local_step
        )

        alpha = lr * ops.sqrt(1 - adam_beta_2_power) / (1 - adam_beta_1_power)

        self.assign_add(
            m, ops.multiply(ops.subtract(gradient, m), 1 - self.adam_beta_1)
        )
        self.assign_add(
            v,
            ops.multiply(
                ops.subtract(ops.square(gradient), v), 1 - self.adam_beta_2
            ),
        )
        self.assign_sub(
            variable,
            ops.divide(
                ops.multiply(m, alpha), ops.add(ops.sqrt(v), self.epsilon)
            ),
        )

    def transpose_last_axis(self, X):
        shape = ops.shape(X)
        temp_order = list(range(len(shape)))
        temp_order[-2] = temp_order[-1]
        temp_order[-1] = len(shape) - 2
        X = ops.transpose(X, temp_order)
        return X

    def lr_adjust(self, x):
        """Adjusts learning rate based on the Moonlight implementation.
        This method enhances the stability of Muon, allowing it to use the same
        learning rate and weight decay as Adam. For details, see
        https://arxiv.org/abs/2502.16982.
        For a 2D matrix, the update is scaled by `sqrt(max(n, m)) * rms_rate`,
        where `n` and `m` are the dimensions of the matrix.
        """
        if self.rms_rate is None:
            return x
        # moonlight version
        # https://github.com/MoonshotAI/Moonlight/blob/master/examples/toy_train.py
        return x * ops.sqrt(ops.maximum(x.shape[0], x.shape[1])) * self.rms_rate

    def zeropower_via_newtonschulz5(self, x, steps: int):
        """We apply the Newton-Schulz iteration to compute matrix G.

        We select a quintic iteration that maximizes the slope at zero. This
        approach helps minimize steps, even if the iteration doesn't fully
        converge across the interval. The result isn't exactly UV^T (from the
        SVD of G), but rather an approximation like US'V^T. Despite this
        approximation, model performance remains unaffected compared to using
        the exact UV^T from the SVD.
        """
        shape = ops.shape(x)
        if len(shape) < 2:
            raise ValueError(
                "Expected gradient or momentum to have at least 2 dimensions. "
                f"Received: shape={shape}"
            )

        a, b, c = self.muon_a, self.muon_b, self.muon_c
        if shape[-2] > shape[-1]:
            x = self.transpose_last_axis(x)

        # Ensure spectral norm is at most 1
        x = x / (ops.norm(x, axis=(-2, -1), keepdims=True) + 1e-7)
        # Perform the NS iterations
        for _ in range(steps):
            temp_a = x @ self.transpose_last_axis(x)
            temp_b = b * temp_a + c * temp_a @ temp_a
            x = a * x + temp_b @ x

        if shape[-2] > shape[-1]:
            x = self.transpose_last_axis(x)
        return x

    def _apply_weight_decay(self, variables):
        for variable in variables:
            if not self._use_weight_decay(variable):
                continue
            if self._should_use_adamw(variable):
                weight_decay_value = self.adam_weight_decay
            else:
                weight_decay_value = self.weight_decay
            if weight_decay_value is None:
                continue
            wd = ops.cast(weight_decay_value, variable.dtype)
            lr = ops.cast(self.learning_rate, variable.dtype)
            variable.assign(variable - variable * wd * lr)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "adam_beta_1": self.adam_beta_1,
                "adam_beta_2": self.adam_beta_2,
                "epsilon": self.epsilon,
                "exclude_layers": self.exclude_layers,
                "muon_a": self.muon_a,
                "muon_b": self.muon_b,
                "muon_c": self.muon_c,
                "adam_lr_ratio": self.adam_lr_ratio,
                "momentum": self.momentum,
                "ns_steps": self.ns_steps,
                "nesterov": self.nesterov,
                "exclude_embeddings": self.exclude_embeddings,
                "adam_weight_decay": self.adam_weight_decay,
                "rms_rate": self.rms_rate,
            }
        )
        return config



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras.src.legacy.preprocessing.image import (
    DirectoryIterator as DirectoryIterator,
)
from keras.src.legacy.preprocessing.image import (
    ImageDataGenerator as ImageDataGenerator,
)
from keras.src.legacy.preprocessing.image import Iterator as Iterator
from keras.src.legacy.preprocessing.image import (
    NumpyArrayIterator as NumpyArrayIterator,
)
from keras.src.legacy.preprocessing.image import (
    apply_affine_transform as apply_affine_transform,
)
from keras.src.legacy.preprocessing.image import (
    apply_brightness_shift as apply_brightness_shift,
)
from keras.src.legacy.preprocessing.image import (
    apply_channel_shift as apply_channel_shift,
)
from keras.src.legacy.preprocessing.image import (
    random_brightness as random_brightness,
)
from keras.src.legacy.preprocessing.image import (
    random_channel_shift as random_channel_shift,
)
from keras.src.legacy.preprocessing.image import (
    random_rotation as random_rotation,
)
from keras.src.legacy.preprocessing.image import random_shear as random_shear
from keras.src.legacy.preprocessing.image import random_shift as random_shift
from keras.src.legacy.preprocessing.image import random_zoom as random_zoom
from keras.src.utils.image_utils import array_to_img as array_to_img
from keras.src.utils.image_utils import img_to_array as img_to_array
from keras.src.utils.image_utils import load_img as load_img
from keras.src.utils.image_utils import save_img as save_img
from keras.src.utils.image_utils import smart_resize as smart_resize



import base64
import io

from packaging.version import parse

from keras.src import backend
from keras.src.api_export import keras_export
from keras.src.layers import Layer
from keras.src.ops import convert_to_numpy
from keras.src.ops import convert_to_tensor
from keras.src.saving.serialization_lib import in_safe_mode


@keras_export("keras.layers.TorchModuleWrapper")
class TorchModuleWrapper(Layer):
    """Torch module wrapper layer.

    `TorchModuleWrapper` is a wrapper class that can turn any
    `torch.nn.Module` into a Keras layer, in particular by making its
    parameters trackable by Keras.

    `TorchModuleWrapper` is only compatible with the PyTorch backend and
    cannot be used with the TensorFlow or JAX backends.

    Args:
        module: `torch.nn.Module` instance. If it's a `LazyModule`
            instance, then its parameters must be initialized before
            passing the instance to `TorchModuleWrapper` (e.g. by calling
            it once).
        output_shape :The shape of the output of this layer. It helps Keras
            perform automatic shape inference.
        name: The name of the layer (string).

    Example:

    Here's an example of how the `TorchModuleWrapper` can be used with vanilla
    PyTorch modules.

    ```python
    import torch
    import torch.nn as nn
    import torch.nn.functional as F

    import keras
    from keras.layers import TorchModuleWrapper

    class Classifier(keras.Model):
        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            # Wrap `torch.nn.Module`s with `TorchModuleWrapper`
            # if they contain parameters
            self.conv1 = TorchModuleWrapper(
                nn.Conv2d(in_channels=1, out_channels=32, kernel_size=(3, 3))
            )
            self.conv2 = TorchModuleWrapper(
                nn.Conv2d(in_channels=32, out_channels=64, kernel_size=(3, 3))
            )
            self.pool = nn.MaxPool2d(kernel_size=(2, 2))
            self.flatten = nn.Flatten()
            self.dropout = nn.Dropout(p=0.5)
            self.fc = TorchModuleWrapper(nn.Linear(1600, 10))

        def call(self, inputs):
            x = F.relu(self.conv1(inputs))
            x = self.pool(x)
            x = F.relu(self.conv2(x))
            x = self.pool(x)
            x = self.flatten(x)
            x = self.dropout(x)
            x = self.fc(x)
            return F.softmax(x, dim=1)


    model = Classifier()
    model.build((1, 28, 28))
    print("Output shape:", model(torch.ones(1, 1, 28, 28).to("cuda")).shape)

    model.compile(
        loss="sparse_categorical_crossentropy",
        optimizer="adam",
        metrics=["accuracy"]
    )
    model.fit(train_loader, epochs=5)
    ```
    """

    def __init__(self, module, name=None, output_shape=None, **kwargs):
        super().__init__(name=name, **kwargs)
        import torch.nn as nn

        from keras.src.backend.torch.core import get_device

        if (
            isinstance(module, nn.modules.lazy.LazyModuleMixin)
            and module.has_uninitialized_params()
        ):
            raise ValueError(
                "LazyModules are not supported unless they "
                "are already initialized. "
                f"Received uninitialized LazyModule: module={module}"
            )

        self.module = module.to(get_device())
        self._track_module_parameters()
        self.output_shape = output_shape

    def parameters(self, recurse=True):
        return self.module.parameters(recurse=recurse)

    def _track_module_parameters(self):
        for param in self.module.parameters():
            # The Variable will reuse the raw `param`
            # and simply wrap it.
            variable = backend.Variable(
                initializer=param, trainable=param.requires_grad
            )
            self._track_variable(variable)
        self.built = True

    def call(self, *args, training=None, **kwargs):
        if training is False:
            self.eval()
        else:
            self.train()
        return self.module(*args, **kwargs)

    def save_own_variables(self, store):
        """Saves model's state from `state_dict`.
        `model.parameters` excludes some of model's state like
        `BatchNorm` mean and variance. So, use `state_dict` to obtain
        all of model's state.
        """
        state_dict = self.module.state_dict()
        for key in state_dict.keys():
            store[key] = convert_to_numpy(state_dict[key])

    def load_own_variables(self, store):
        """Loads model's state via `state_dict`."""
        state_dict = {}
        for key in store.keys():
            if isinstance(key, bytes):
                key = key.decode()
            state_dict[key] = convert_to_tensor(store[key])
        self.module.load_state_dict(state_dict)

    def compute_output_shape(self, input_shape):
        if self.output_shape is None:
            return super().compute_output_shape(input_shape)
        return self.output_shape

    def get_config(self):
        base_config = super().get_config()
        import torch

        buffer = io.BytesIO()
        torch.save(self.module, buffer)
        # Encode the buffer using base64 to ensure safe serialization
        buffer_b64 = base64.b64encode(buffer.getvalue()).decode("ascii")
        config = {
            "module": buffer_b64,
            "output_shape": self.output_shape,
        }
        return {**base_config, **config}

    @classmethod
    def from_config(cls, config):
        import torch

        if "module" in config:
            if in_safe_mode():
                raise ValueError(
                    "Requested the deserialization of a `torch.nn.Module` "
                    "object via `torch.load()`. This carries a potential risk "
                    "of arbitrary code execution and thus it is disallowed by "
                    "default. If you trust the source of the artifact, you can "
                    "override this error by passing `safe_mode=False` to the "
                    "loading function, or calling "
                    "`keras.config.enable_unsafe_deserialization()."
                )

            # Decode the base64 string back to bytes
            buffer_bytes = base64.b64decode(config["module"].encode("ascii"))
            buffer = io.BytesIO(buffer_bytes)
            config["module"] = torch.load(buffer, weights_only=False)
        return cls(**config)


def no_grad(orig_func):
    import torch

    if parse(torch.__version__) >= parse("2.1.0"):
        return torch.no_grad(orig_func)
    else:
        return orig_func



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras.src.initializers import deserialize as deserialize
from keras.src.initializers import get as get
from keras.src.initializers import serialize as serialize
from keras.src.initializers.constant_initializers import STFT as STFT
from keras.src.initializers.constant_initializers import STFT as STFTInitializer
from keras.src.initializers.constant_initializers import STFT as stft
from keras.src.initializers.constant_initializers import Constant as Constant
from keras.src.initializers.constant_initializers import Constant as constant
from keras.src.initializers.constant_initializers import Identity as Identity
from keras.src.initializers.constant_initializers import (
    Identity as IdentityInitializer,
)
from keras.src.initializers.constant_initializers import Identity as identity
from keras.src.initializers.constant_initializers import Ones as Ones
from keras.src.initializers.constant_initializers import Ones as ones
from keras.src.initializers.constant_initializers import Zeros as Zeros
from keras.src.initializers.constant_initializers import Zeros as zeros
from keras.src.initializers.initializer import Initializer as Initializer
from keras.src.initializers.random_initializers import (
    GlorotNormal as GlorotNormal,
)
from keras.src.initializers.random_initializers import (
    GlorotNormal as glorot_normal,
)
from keras.src.initializers.random_initializers import (
    GlorotUniform as GlorotUniform,
)
from keras.src.initializers.random_initializers import (
    GlorotUniform as glorot_uniform,
)
from keras.src.initializers.random_initializers import HeNormal as HeNormal
from keras.src.initializers.random_initializers import HeNormal as he_normal
from keras.src.initializers.random_initializers import HeUniform as HeUniform
from keras.src.initializers.random_initializers import HeUniform as he_uniform
from keras.src.initializers.random_initializers import (
    LecunNormal as LecunNormal,
)
from keras.src.initializers.random_initializers import (
    LecunNormal as lecun_normal,
)
from keras.src.initializers.random_initializers import (
    LecunUniform as LecunUniform,
)
from keras.src.initializers.random_initializers import (
    LecunUniform as lecun_uniform,
)
from keras.src.initializers.random_initializers import Orthogonal as Orthogonal
from keras.src.initializers.random_initializers import (
    Orthogonal as OrthogonalInitializer,
)
from keras.src.initializers.random_initializers import Orthogonal as orthogonal
from keras.src.initializers.random_initializers import (
    RandomNormal as RandomNormal,
)
from keras.src.initializers.random_initializers import (
    RandomNormal as random_normal,
)
from keras.src.initializers.random_initializers import (
    RandomUniform as RandomUniform,
)
from keras.src.initializers.random_initializers import (
    RandomUniform as random_uniform,
)
from keras.src.initializers.random_initializers import (
    TruncatedNormal as TruncatedNormal,
)
from keras.src.initializers.random_initializers import (
    TruncatedNormal as truncated_normal,
)
from keras.src.initializers.random_initializers import (
    VarianceScaling as VarianceScaling,
)
from keras.src.initializers.random_initializers import (
    VarianceScaling as variance_scaling,
)



from keras.src import backend
from keras.src.api_export import keras_export
from keras.src.layers.preprocessing.image_preprocessing.base_image_preprocessing_layer import (  # noqa: E501
    BaseImagePreprocessingLayer,
)
from keras.src.layers.preprocessing.image_preprocessing.bounding_boxes.converters import (  # noqa: E501
    clip_to_image_size,
)
from keras.src.layers.preprocessing.image_preprocessing.bounding_boxes.converters import (  # noqa: E501
    convert_format,
)
from keras.src.ops.core import _saturate_cast


@keras_export("keras.layers.Resizing")
class Resizing(BaseImagePreprocessingLayer):
    """A preprocessing layer which resizes images.

    This layer resizes an image input to a target height and width. The input
    should be a 4D (batched) or 3D (unbatched) tensor in `"channels_last"`
    format. Input pixel values can be of any range
    (e.g. `[0., 1.)` or `[0, 255]`).

    **Note:** This layer is safe to use inside a `tf.data` or `grain` pipeline
    (independently of which backend you're using).

    Input shape:
        3D (unbatched) or 4D (batched) tensor with shape:
        `(..., height, width, channels)`, in `"channels_last"` format,
        or `(..., channels, height, width)`, in `"channels_first"` format.

    Output shape:
        3D (unbatched) or 4D (batched) tensor with shape:
        `(..., target_height, target_width, channels)`,
        or `(..., channels, target_height, target_width)`,
        in `"channels_first"` format.

    Args:
        height: Integer, the height of the output shape.
        width: Integer, the width of the output shape.
        interpolation: String, the interpolation method.
            Supports `"bilinear"`, `"nearest"`, `"bicubic"`,
            `"lanczos3"`, `"lanczos5"`. Defaults to `"bilinear"`.
        crop_to_aspect_ratio: If `True`, resize the images without aspect
            ratio distortion. When the original aspect ratio differs
            from the target aspect ratio, the output image will be
            cropped so as to return the
            largest possible window in the image (of size `(height, width)`)
            that matches the target aspect ratio. By default
            (`crop_to_aspect_ratio=False`), aspect ratio may not be preserved.
        pad_to_aspect_ratio: If `True`, pad the images without aspect
            ratio distortion. When the original aspect ratio differs
            from the target aspect ratio, the output image will be
            evenly padded on the short side.
        fill_mode: When using `pad_to_aspect_ratio=True`, padded areas
            are filled according to the given mode. Only `"constant"` is
            supported at this time
            (fill with constant value, equal to `fill_value`).
        fill_value: Float. Padding value to use when `pad_to_aspect_ratio=True`.
        data_format: string, either `"channels_last"` or `"channels_first"`.
            The ordering of the dimensions in the inputs. `"channels_last"`
            corresponds to inputs with shape `(batch, height, width, channels)`
            while `"channels_first"` corresponds to inputs with shape
            `(batch, channels, height, width)`. It defaults to the
            `image_data_format` value found in your Keras config file at
            `~/.keras/keras.json`. If you never set it, then it will be
            `"channels_last"`.
        **kwargs: Base layer keyword arguments, such as `name` and `dtype`.

    Example:

    ```python
    (x_train, y_train), _ = keras.datasets.cifar10.load_data()
    image = x_train[0]
    resizer = keras.layers.Resizing(128, 128)
    resized_image = resizer(image)
    print("original:", image.shape, "resized:", resized_image.shape)
    ```
    """

    _USE_BASE_FACTOR = False

    def __init__(
        self,
        height,
        width,
        interpolation="bilinear",
        crop_to_aspect_ratio=False,
        pad_to_aspect_ratio=False,
        fill_mode="constant",
        fill_value=0.0,
        antialias=False,
        data_format=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.height = height
        self.width = width
        self.interpolation = interpolation
        self.data_format = backend.standardize_data_format(data_format)
        self.crop_to_aspect_ratio = crop_to_aspect_ratio
        self.pad_to_aspect_ratio = pad_to_aspect_ratio
        self.fill_mode = fill_mode
        self.fill_value = fill_value
        self.antialias = bool(antialias)
        if self.data_format == "channels_first":
            self.height_axis = -2
            self.width_axis = -1
        elif self.data_format == "channels_last":
            self.height_axis = -3
            self.width_axis = -2

    def _transform_images(self, images, transformation, interpolation):
        size = (self.height, self.width)
        resized = self.backend.image.resize(
            images,
            size=size,
            interpolation=interpolation,
            antialias=self.antialias,
            data_format=self.data_format,
            crop_to_aspect_ratio=self.crop_to_aspect_ratio,
            pad_to_aspect_ratio=self.pad_to_aspect_ratio,
            fill_mode=self.fill_mode,
            fill_value=self.fill_value,
        )
        if resized.dtype == images.dtype:
            return resized
        if backend.is_int_dtype(images.dtype):
            resized = self.backend.numpy.round(resized)
        return _saturate_cast(resized, images.dtype, self.backend)

    def transform_images(self, images, transformation=None, training=True):
        return self._transform_images(
            images, transformation, self.interpolation
        )

    def transform_segmentation_masks(
        self, segmentation_masks, transformation=None, training=True
    ):
        return self._transform_images(
            segmentation_masks, transformation, "nearest"
        )

    def transform_labels(self, labels, transformation=None, training=True):
        return labels

    def get_random_transformation(self, data, training=True, seed=None):
        if isinstance(data, dict):
            input_shape = self.backend.shape(data["images"])
        else:
            input_shape = self.backend.shape(data)

        input_height, input_width = (
            input_shape[self.height_axis],
            input_shape[self.width_axis],
        )

        return input_height, input_width

    def transform_bounding_boxes(
        self,
        bounding_boxes,
        transformation,
        training=True,
    ):
        ops = self.backend
        input_height, input_width = transformation
        mask_negative_1s = ops.numpy.all(bounding_boxes["boxes"] == -1, axis=-1)
        mask_zeros = ops.numpy.all(bounding_boxes["boxes"] == 0, axis=-1)
        boxes_mask = ops.numpy.logical_or(mask_negative_1s, mask_zeros)

        bounding_boxes = convert_format(
            bounding_boxes,
            source=self.bounding_box_format,
            target="xyxy",
            height=input_height,
            width=input_width,
        )

        bounding_boxes["boxes"] = self._transform_xyxy(
            bounding_boxes["boxes"],
            input_height=input_height,
            input_width=input_width,
        )

        bounding_boxes = clip_to_image_size(
            bounding_boxes=bounding_boxes,
            height=self.height,
            width=self.width,
        )

        bounding_boxes["boxes"] = ops.numpy.where(
            ops.numpy.expand_dims(boxes_mask, axis=-1),
            ops.convert_to_tensor(
                [0.0, 0.0, 0.0, 0.0], dtype=bounding_boxes["boxes"].dtype
            ),
            bounding_boxes["boxes"],
        )

        bounding_boxes = convert_format(
            bounding_boxes,
            source="xyxy",
            target=self.bounding_box_format,
            height=self.height,
            width=self.width,
        )

        return bounding_boxes

    def _transform_xyxy(self, boxes, input_height, input_width):
        ops = self.backend
        input_height = ops.cast(input_height, dtype=boxes.dtype)
        input_width = ops.cast(input_width, dtype=boxes.dtype)

        if self.pad_to_aspect_ratio:
            return self._transform_boxes_pad_to_aspect_ratio(
                boxes, input_height, input_width
            )
        elif self.crop_to_aspect_ratio:
            return self._transform_boxes_crop_to_aspect_ratio(
                boxes, input_height, input_width
            )
        else:
            return self._transform_boxes_stretch(
                boxes, input_height, input_width
            )

    def _transform_boxes_pad_to_aspect_ratio(
        self, boxes, input_height, input_width
    ):
        """Transforms bounding boxes for padding to aspect ratio."""
        ops = self.backend
        height_ratio = ops.cast(self.height / input_height, dtype=boxes.dtype)
        width_ratio = ops.cast(self.width / input_width, dtype=boxes.dtype)
        min_aspect_ratio = ops.numpy.minimum(height_ratio, width_ratio)
        y_offset = (self.height - input_height * min_aspect_ratio) // 2
        x_offset = (self.width - input_width * min_aspect_ratio) // 2
        return ops.numpy.stack(
            [
                boxes[..., 0] * min_aspect_ratio + x_offset,
                boxes[..., 1] * min_aspect_ratio + y_offset,
                boxes[..., 2] * min_aspect_ratio + x_offset,
                boxes[..., 3] * min_aspect_ratio + y_offset,
            ],
            axis=-1,
        )

    def _transform_boxes_crop_to_aspect_ratio(
        self, boxes, input_height, input_width
    ):
        """Transforms bounding boxes for cropping to aspect ratio."""
        ops = self.backend
        source_aspect_ratio = input_width / input_height
        target_aspect_ratio = self.width / self.height
        new_width = ops.numpy.where(
            source_aspect_ratio > target_aspect_ratio,
            self.height * source_aspect_ratio,
            self.width,
        )
        new_height = ops.numpy.where(
            source_aspect_ratio > target_aspect_ratio,
            self.height,
            self.width / source_aspect_ratio,
        )
        scale_x = new_width / input_width
        scale_y = new_height / input_height
        crop_left = (new_width - self.width) // 2
        crop_top = (new_height - self.height) // 2
        return ops.numpy.stack(
            [
                boxes[..., 0] * scale_x - crop_left,
                boxes[..., 1] * scale_y - crop_top,
                boxes[..., 2] * scale_x - crop_left,
                boxes[..., 3] * scale_y - crop_top,
            ],
            axis=-1,
        )

    def _transform_boxes_stretch(self, boxes, input_height, input_width):
        """Transforms bounding boxes by simple stretching."""
        ops = self.backend
        height_ratio = ops.cast(self.height / input_height, dtype=boxes.dtype)
        width_ratio = ops.cast(self.width / input_width, dtype=boxes.dtype)
        return ops.numpy.stack(
            [
                boxes[..., 0] * width_ratio,
                boxes[..., 1] * height_ratio,
                boxes[..., 2] * width_ratio,
                boxes[..., 3] * height_ratio,
            ],
            axis=-1,
        )

    def compute_output_shape(self, input_shape):
        input_shape = list(input_shape)
        if len(input_shape) == 4:
            if self.data_format == "channels_last":
                input_shape[1] = self.height
                input_shape[2] = self.width
            else:
                input_shape[2] = self.height
                input_shape[3] = self.width
        else:
            if self.data_format == "channels_last":
                input_shape[0] = self.height
                input_shape[1] = self.width
            else:
                input_shape[1] = self.height
                input_shape[2] = self.width
        return tuple(input_shape)

    def get_config(self):
        base_config = super().get_config()
        config = {
            "height": self.height,
            "width": self.width,
            "interpolation": self.interpolation,
            "crop_to_aspect_ratio": self.crop_to_aspect_ratio,
            "pad_to_aspect_ratio": self.pad_to_aspect_ratio,
            "fill_mode": self.fill_mode,
            "fill_value": self.fill_value,
            "antialias": self.antialias,
            "data_format": self.data_format,
        }
        return {**base_config, **config}
